#line 1 "-"
#line 1 "/home/lagavulin/a/gniady/rsim-1.0/apps/utils/lib/rsim_macros"

#line 120

#line 1 "src/code.C"
/*************************************************************************/
/*                                                                       */
/*  Copyright (c) 1994 Stanford University                               */
/*                                                                       */
/*  All rights reserved.                                                 */
/*                                                                       */
/*  Permission is given to use, copy, and modify this software for any   */
/*  non-commercial purpose as long as this copyright notice is not       */
/*  removed.  All other uses, including redistribution in whole or in    */
/*  part, are forbidden without prior written permission.                */
/*                                                                       */
/*  This software is provided with absolutely no warranty and no         */
/*  support.                                                             */
/*                                                                       */
/*************************************************************************/
/*
   Usage: barnes <options> < inputfile
   
   Command line options:
   
   -h : Print out input file description
   
   Input parameters should be placed in a file and redirected through
   standard input.  There are a total of twelve parameters, and all of 
   them have default values.
   
   1) infile (char*) : The name of an input file that contains particle
   data.
   
   The format of the file is:
   a) An int representing the number of particles in the distribution
   b) An int representing the dimensionality of the problem (3-D)
   c) A double representing the current time of the simulation
   d) Doubles representing the masses of all the particles
   e) A vector (length equal to the dimensionality) of doubles
   representing the positions of all the particles
   f) A vector (length equal to the dimensionality) of doubles
   representing the velocities of all the particles
   
   Each of these numbers can be separated by any amount of whitespace.
   2) nbody (int) : If no input file is specified (the first line is
   blank), this number specifies the number of particles to generate
   under a plummer model.  Default is 16384.
   3) seed (int) : The seed used by the random number generator.
   Default is 123.
   4) outfile (char*) : The name of the file that snapshots will be
   printed to. This feature has been disabled in the SPLASH release.
   Default is NULL.
   5) dtime (double) : The integration time-step.
   Default is 0.025.
   6) eps (double) : The usual potential softening
   Default is 0.05.
   7) tol (double) : The cell subdivision tolerance.
   Default is 1.0.
   8) fcells (double) : Number of cells created = fcells * number of 
   leaves.
   Default is 2.0.
   9) fleaves (double) : Number of leaves created = fleaves * nbody.
   Default is 0.5.
   10) tstop (double) : The time to stop integration.
   Default is 0.075.
   11) dtout (double) : The data-output interval.
   Default is 0.25.
   */

#ifdef WWT
{;} 
#else
#include <rsim_apps.h>
TreeBar final_sync;
#line 70
#endif

#ifdef RSIM
int phase = 0;
#define MASTER 0
#endif

#define PUBLIC /* nada */

#include <math.h>
#include <time.h>
#include <stdlib.h>
#include "code.h"
#include "defs.h"

#ifdef BLZACC_hwstache
#include "hwstache.h"
#endif

static int Child_Sequence[NUM_DIRECTIONS][NSUB] =
{
  { 2, 5, 6, 1, 0, 3, 4, 7},  /* BRC_FUC */
  { 2, 5, 6, 1, 0, 7, 4, 3},  /* BRC_FRA */
  { 1, 6, 5, 2, 3, 0, 7, 4},  /* BRA_FDA */
  { 1, 6, 5, 2, 3, 4, 7, 0},  /* BRA_FRC */
  { 6, 1, 2, 5, 4, 7, 0, 3},  /* BLC_FDC */
  { 6, 1, 2, 5, 4, 3, 0, 7},  /* BLC_FLA */
  { 5, 2, 1, 6, 7, 4, 3, 0},  /* BLA_FUA */
  { 5, 2, 1, 6, 7, 0, 3, 4},  /* BLA_FLC */
  { 1, 2, 5, 6, 7, 4, 3, 0},  /* BUC_FUA */
  { 1, 2, 5, 6, 7, 0, 3, 4},  /* BUC_FLC */
  { 6, 5, 2, 1, 0, 3, 4, 7},  /* BUA_FUC */
  { 6, 5, 2, 1, 0, 7, 4, 3},  /* BUA_FRA */
  { 5, 6, 1, 2, 3, 0, 7, 4},  /* BDC_FDA */
  { 5, 6, 1, 2, 3, 4, 7, 0},  /* BDC_FRC */
  { 2, 1, 6, 5, 4, 7, 0, 3},  /* BDA_FDC */
  { 2, 1, 6, 5, 4, 3, 0, 7},  /* BDA_FLA */
    
  { 3, 4, 7, 0, 1, 2, 5, 6},  /* FRC_BUC */
  { 3, 4, 7, 0, 1, 6, 5, 2},  /* FRC_BRA */
  { 0, 7, 4, 3, 2, 1, 6, 5},  /* FRA_BDA */
  { 0, 7, 4, 3, 2, 5, 6, 1},  /* FRA_BRC */
  { 7, 0, 3, 4, 5, 6, 1, 2},  /* FLC_BDC */
  { 7, 0, 3, 4, 5, 2, 1, 6},  /* FLC_BLA */
  { 4, 3, 0, 7, 6, 5, 2, 1},  /* FLA_BUA */
  { 4, 3, 0, 7, 6, 1, 2, 5},  /* FLA_BLC */
  { 0, 3, 4, 7, 6, 5, 2, 1},  /* FUC_BUA */
  { 0, 3, 4, 7, 6, 1, 2, 5},  /* FUC_BLC */
  { 7, 4, 3, 0, 1, 2, 5, 6},  /* FUA_BUC */
  { 7, 4, 3, 0, 1, 6, 5, 2},  /* FUA_BRA */
  { 4, 7, 0, 3, 2, 1, 6, 5},  /* FDC_BDA */
  { 4, 7, 0, 3, 2, 5, 6, 1},  /* FDC_BRC */
  { 3, 0, 7, 4, 5, 6, 1, 2},  /* FDA_BDC */
  { 3, 0, 7, 4, 5, 2, 1, 6},  /* FDA_BLA */
};

static int Direction_Sequence[NUM_DIRECTIONS][NSUB] =
{
  { FRC_BUC, BRA_FRC, FDA_BDC, BLA_FUA, BUC_FLC, FUA_BUC, BRA_FRC, FDA_BLA },
    /* BRC_FUC */
  { FRC_BUC, BRA_FRC, FDA_BDC, BLA_FUA, BRA_FDA, FRC_BRA, BUC_FUA, FLC_BDC },
    /* BRC_FRA */
  { FRA_BDA, BRC_FRA, FUC_BUA, BLC_FDC, BDA_FLA, FDC_BDA, BRC_FRA, FUC_BLC },
    /* BRA_FDA */
  { FRA_BDA, BRC_FRA, FUC_BUA, BLC_FDC, BUC_FLC, FUA_BUC, BRA_FRC, FDA_BLA },
    /* BRA_FRC */
  { FLC_BDC, BLA_FLC, FUA_BUC, BRA_FDA, BDC_FRC, FDA_BDC, BLA_FLC, FUA_BRA },
    /* BLC_FDC */
  { FLC_BDC, BLA_FLC, FUA_BUC, BRA_FDA, BLA_FUA, FLC_BLA, BDC_FDA, FRC_BUC },
    /* BLC_FLA */
  { FLA_BUA, BLC_FLA, FDC_BDA, BRC_FUC, BUA_FRA, FUC_BUA, BLC_FLA, FDC_BRC },
    /* BLA_FUA */
  { FLA_BUA, BLC_FLA, FDC_BDA, BRC_FUC, BLC_FDC, FLA_BLC, BUA_FUC, FRA_BDA },
    /* BLA_FLC */
  { FUC_BLC, BUA_FUC, FRA_BRC, BDA_FLA, BUA_FRA, FUC_BUA, BLC_FLA, FDC_BRC },
    /* BUC_FUA */
  { FUC_BLC, BUA_FUC, FRA_BRC, BDA_FLA, BLC_FDC, FLA_BLC, BUA_FUC, FRA_BDA },
    /* BUC_FLC */
  { FUA_BRA, BUC_FUA, FLC_BLA, BDC_FRC, BUC_FLC, FUA_BUC, BRA_FRC, FDA_BLA },
    /* BUA_FUC */
  { FUA_BRA, BUC_FUA, FLC_BLA, BDC_FRC, BRA_FDA, FRC_BRA, BUC_FUA, FLC_BDC },
    /* BUA_FRA */
  { FDC_BRC, BDA_FDC, FLA_BLC, BUA_FRA, BDA_FLA, FDC_BDA, BRC_FRA, FUC_BLC },
    /* BDC_FDA */
  { FDC_BRC, BDA_FDC, FLA_BLC, BUA_FRA, BUC_FLC, FUA_BUC, BRA_FRC, FDA_BLA },
    /* BDC_FRC */
  { FDA_BLA, BDC_FDA, FRC_BRA, BUC_FLC, BDC_FRC, FDA_BDC, BLA_FLC, FUA_BRA },
    /* BDA_FDC */
  { FDA_BLA, BDC_FDA, FRC_BRA, BUC_FLC, BLA_FUA, FLC_BLA, BDC_FDA, FRC_BUC },
    /* BDA_FLA */
    
  { BUC_FLC, FUA_BUC, BRA_FRC, FDA_BLA, FUC_BLC, BUA_FUC, FRA_BRC, BDA_FLA },
    /* FRC_BUC */
  { BUC_FLC, FUA_BUC, BRA_FRC, FDA_BLA, FRA_BDA, BRC_FRA, FUC_BUA, BLC_FDC },
    /* FRC_BRA */
  { BRA_FDA, FRC_BRA, BUC_FUA, FLC_BDC, FDA_BLA, BDC_FDA, FRC_BRA, BUC_FLC },
    /* FRA_BDA */
  { BRA_FDA, FRC_BRA, BUC_FUA, FLC_BDC, FRC_BUC, BRA_FRC, FDA_BDC, BLA_FUA },
    /* FRA_BRC */
  { BLC_FDC, FLA_BLC, BUA_FUC, FRA_BDA, FDC_BRC, BDA_FDC, FLA_BLC, BUA_FRA },
    /* FLC_BDC */
  { BLC_FDC, FLA_BLC, BUA_FUC, FRA_BDA, FLA_BUA, BLC_FLA, FDC_BDA, BRC_FUC },
    /* FLC_BLA */
  { BLA_FUA, FLC_BLA, BDC_FDA, FRC_BUC, FUA_BRA, BUC_FUA, FLC_BLA, BDC_FRC },
    /* FLA_BUA */
  { BLA_FUA, FLC_BLA, BDC_FDA, FRC_BUC, FLC_BDC, BLA_FLC, FUA_BUC, BRA_FDA },
    /* FLA_BLC */
  { BUC_FLC, FUA_BUC, BRA_FRC, FDA_BLA, FUA_BRA, BUC_FUA, FLC_BLA, BDC_FRC },
    /* FUC_BUA */
  { BUC_FLC, FUA_BUC, BRA_FRC, FDA_BLA, FLC_BDC, BLA_FLC, FUA_BUC, BRA_FDA },
    /* FUC_BLC */
  { BUA_FRA, FUC_BUA, BLC_FLA, FDC_BRC, FUC_BLC, BUA_FUC, FRA_BRC, BDA_FLA },
    /* FUA_BUC */
  { BUA_FRA, FUC_BUA, BLC_FLA, FDC_BRC, FRA_BDA, BRC_FRA, FUC_BUA, BLC_FDC },
    /* FUA_BRA */
  { BDC_FRC, FDA_BDC, BLA_FLC, FUA_BRA, FDA_BLA, BDC_FDA, FRC_BRA, BUC_FLC },
    /* FDC_BDA */
  { BDC_FRC, FDA_BDC, BLA_FLC, FUA_BRA, FRC_BUC, BRA_FRC, FDA_BDC, BLA_FUA },
    /* FDC_BRC */
  { BDA_FLA, FDC_BDA, BRC_FRA, FUC_BLC, FDC_BRC, BDA_FDC, FLA_BLC, BUA_FRA },
    /* FDA_BDC */
  { BDA_FLA, FDC_BDA, BRC_FRA, FUC_BLC, FLA_BUA, BLC_FLA, FDC_BDA, BRC_FUC },
    /* FDA_BLA */
};

string defv[] = {                 /* DEFAULT PARAMETER VALUES              */
    /* file names for input/output                                         */
    "in=",                        /* snapshot of initial conditions        */
    "out=",                       /* stream of output snapshots            */
    
    /* params, used if no input specified, to make a Plummer Model         */
    "nbody=256",                  /* number of particles to generate       */
    "seed=123",                   /* random number generator seed          */
    
    /* params to control N-body integration                                */
    "dtime=0.025",                /* integration time-step                 */
    "eps=0.05",                   /* usual potential softening             */
    "tol=1.0",                    /* cell subdivision tolerence            */
    "fcells=2.0",                 /* cell allocation parameter             */
    "fleaves=0.5",                /* leaf allocation parameter             */
    
    "tstop=0.075",                 /* time to stop integration             */
    "dtout=0.25",                  /* data-output interval                 */
};

void SlaveStart ();
void stepsystem (unsigned int ThreadId);
void ComputeForces ();
void Help();
FILE *fopen();

void ANLinit();
void initparam(string argv[], string defv[]);
void startrun();
void initoutput();
void tab_init();

unsigned long long starttime, inittime;

void
#if defined(BLZ) || defined(TZERO)
real_main(argc, argv)
#else
main(argc, argv)
#endif BLZ
     int argc;
     string argv[];
{
#ifdef RSIM
MEMSYS_OFF
#endif
    if (argc != 2)
    {
	fprintf(stderr, "usage: %s\n", argv[0]);
	fflush(stderr);
	Help();
	exit(-1);
    }
    
    NTHREAD = atoi(argv[1]);
    
    {long time(long *); (starttime) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;

    ANLinit();
    initparam(argv, defv);
    startrun();
    initoutput();
    tab_init();
    
    Global->tracktime = 0;
    Global->partitiontime = 0;
    Global->treebuildtime = 0;
    Global->forcecalctime = 0;
    
    /* Create the slave processes: number of threads less one,
       since the master will do work as well */
    Global->current_id = 0;
    
#ifdef NO_FORK_SEMANTICS
    set_fork_semantics(&Global, sizeof(Global));
    set_fork_semantics(&dtime, sizeof(dtime));
    set_fork_semantics(&dtout, sizeof(dtout));
    set_fork_semantics(&tstop, sizeof(tstop));
    set_fork_semantics(&nbody, sizeof(nbody));
    set_fork_semantics(&fcells, sizeof(fcells));
    set_fork_semantics(&fleaves, sizeof(fleaves));
    set_fork_semantics(&tol, sizeof(tol));
    set_fork_semantics(&tolsq, sizeof(tolsq));
    set_fork_semantics(&eps, sizeof(eps));
    set_fork_semantics(&epssq, sizeof(epssq));
    set_fork_semantics(&dthf, sizeof(dthf));
    set_fork_semantics(&NTHREAD, sizeof(NTHREAD));
    set_fork_semantics(&maxcell, sizeof(maxcell));
    set_fork_semantics(&maxleaf, sizeof(maxleaf));
    set_fork_semantics(&maxmybody, sizeof(maxmybody));
    set_fork_semantics(&maxmycell, sizeof(maxmycell));
    set_fork_semantics(&maxmyleaf, sizeof(maxmyleaf));
    set_fork_semantics(&bodytab, sizeof(bodytab));
    set_fork_semantics(&CellLock, sizeof(CellLock));
    set_fork_semantics(&Local[0], sizeof(struct local_memory));
  {
      int ThreadId;
      
      for (ThreadId = 1; ThreadId < NTHREAD; ++ThreadId)
      {
	  set_fork_semantics(&Local[ThreadId].ctab, 
			     sizeof(Local[ThreadId].ctab));
	  set_fork_semantics(&Local[ThreadId].ltab, 
			     sizeof(Local[ThreadId].ltab));
      }
  }
#endif NO_FORK_SEMANTICS
    
#ifdef RSIM
MEMSYS_ON;
#endif    

#ifdef WWT
    CREATE_ALL(SlaveStart);
#else
  {
      unsigned ThreadId;
      for(ThreadId = 1; ThreadId < NTHREAD; ThreadId++)
      {
	  { if(fork() == 0) {(SlaveStart)(); TREEBAR(&final_sync, getpid()); exit(0);}};
      }
  }
#endif
    
    printf("COMPUTESTART  = %12.0f\n", (double)Global->computestart);
    
    SlaveStart();
    
    {long time(long *); (Global->computeend) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    
    {TREEBAR(&final_sync, getpid());};

    printf("COMPUTEEND    = %12.0f\n", (double)Global->computeend);
    printf("COMPUTETIME   = %12.0f\n", (double)Global->computeend - Global->computestart);
    printf("TRACKTIME     = %12.0f\n", (double)Global->tracktime); 
    printf("PARTITIONTIME = %12.0f\t%5.2f\n", 
	   (double)Global->partitiontime,
	   ((double)Global->partitiontime)/(double)Global->tracktime);
    printf("TREEBUILDTIME = %12.0f\t%5.2f\n", 
	   (double)Global->treebuildtime, 
	   ((double)Global->treebuildtime)/(double)Global->tracktime);
    printf("FORCECALCTIME = %12.0f\t%5.2f\n", 
	   (double)Global->forcecalctime,
	   ((double)Global->forcecalctime)/(double)Global->tracktime);
    printf("RESTTIME      = %12.0f\t%5.2f\n", 
	   (double) (Global->tracktime - Global->partitiontime - 
		     Global->treebuildtime - Global->forcecalctime), 
	   ((double)(Global->tracktime-Global->partitiontime-
		     Global->treebuildtime-Global->forcecalctime))/
	   (double)Global->tracktime);

    printf("Initialization = %1.0f\n", (double)inittime);
    printf("Computation = %1.0f\n", (double)Global->tracktime);


    {exit(0);};
}

/*
 * ANLINIT : initialize ANL macros
 */
void
ANLinit()
{
#ifdef WWT    
    INITENV(80000000);
#else
    {{TreeBarInit(&(final_sync), NUMBER_OF_PROCESSORS);};};
#endif
    /* Allocate global, shared memory */
    fprintf(stderr,"Allocating Global Memory, size: %d \n",
	    sizeof(struct GlobalMemory));
    
    fflush(stderr);
    Global = (struct GlobalMemory *)shmalloc(sizeof(struct GlobalMemory));;
    if (Global==NULL) error("No initialization for Global\n");
    
    {TreeBarInit(&(Global->Barload), NUMBER_OF_PROCESSORS);}
    {TreeBarInit(&(Global->Bartree), NUMBER_OF_PROCESSORS);}
    {TreeBarInit(&(Global->Barcom), NUMBER_OF_PROCESSORS);}
    {TreeBarInit(&(Global->Baraccel), NUMBER_OF_PROCESSORS);}
    {TreeBarInit(&(Global->Barstart), NUMBER_OF_PROCESSORS);}
    {TreeBarInit(&(Global->Barpos), NUMBER_OF_PROCESSORS);}
    {TreeBarInit(&(Global->Barover), NUMBER_OF_PROCESSORS);}
    {FREELOCK(&Global->CountLock);};
    {FREELOCK(&Global->io_lock);};
}

/*
 * INIT_ROOT: Thread 0 reinitialize the global root at each time step
 */
void
init_root (ThreadId)
     unsigned int ThreadId;
{
    int i;
    
    Global->G_root=Local[0].ctab;
    Type(Global->G_root) = CELL;
    {WRITEFLAG(&(Done(Global->G_root)),0);};
    Level(Global->G_root) = IMAX >> 1;
    for (i = 0; i < NSUB; i++) 
    {
	Subp(Global->G_root)[i] = NULL;
    }
    Local[0].mynumcell=1;
}

int Log_base_2(number)
     int number;
{
    int cumulative;
    int out;
    
    cumulative = 1;
    for (out = 0; out < 20; out++) 
    {
	if (cumulative == number) 
	{
	    return(out);
	}
	else 
	{
	    cumulative = cumulative * 2;
	}
    }
    
    fprintf(stderr,"Log_base_2: couldn't find log2 of %d\n", number);
    fflush(stderr);
    exit(-1);
}

/*
 * TAB_INIT : allocate body and cell data space
 */
void
tab_init()
{
    int i;
    
    /*allocate leaf/cell space */
    maxleaf = (int) ((double) fleaves * nbody);
    maxcell = fcells * maxleaf;
    for (i = 0; i < NTHREAD; ++i) 
    {
	Local[i].ctab = (cellptr) shmalloc((maxcell / NTHREAD) * sizeof(cell));;
	Local[i].ltab = (leafptr) shmalloc((maxleaf / NTHREAD) * sizeof(leaf));;
    }
    
    /*allocate space for personal lists of body pointers */
    maxmybody = (nbody+maxleaf*MAX_BODIES_PER_LEAF)/NTHREAD; 
    Local[0].mybodytab = (bodyptr*)shmalloc(NTHREAD*maxmybody*sizeof(bodyptr));;
    /* space is allocated so that every */
    /* process can have a maximum of maxmybody pointers to bodies */ 
    /* then there is an array of bodies called bodytab which is  */
    /* allocated in the distribution generation or when the distr. */
    /* file is read */
    maxmycell = maxcell / NTHREAD;
    maxmyleaf = maxleaf / NTHREAD;
    Local[0].mycelltab = (cellptr*)shmalloc(NTHREAD*maxmycell*sizeof(cellptr));;
    Local[0].myleaftab = (leafptr*)shmalloc(NTHREAD*maxmyleaf*sizeof(leafptr));;
    
    CellLock = (struct CellLockType *) shmalloc(sizeof(struct CellLockType));;
    {
#line 458
  int tempk;
#line 458
  for(tempk=0;tempk<MAXLOCK;tempk++)
#line 458
     FREELOCK(&(CellLock->CL[tempk]));
#line 458
};
}

/*
 * SLAVESTART: main task for each thread
 */
void find_my_initial_bodies(bodyptr btab, int nbody, unsigned int ThreadId);

void 
SlaveStart()
{
    int iteration = 0;
    unsigned int ThreadId;

    ThreadId = getpid();

    if (ThreadId == 0)
    {
	/* Make the master do slave work so we don't waste the thread */
	{long time(long *); (inittime) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
	inittime -= starttime;
    }

    {TREEBAR(&(Global->Barstart), MyNum);};

#ifdef BLZACC_hwstache
    if (threadId() == 0)
    {
	mark_pages_for_migration();
    }
    threadGlobalBarrier();
#endif

    if (ThreadId == 0)
    {
	{long time(long *); (Global->computestart) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    }

    printf("Thread %d running\n", ThreadId);
    
    /* POSSIBLE ENHANCEMENT:  Here is where one might pin processes to
       threads to avoid migration */
    
    /* initialize mybodytabs */
    Local[ThreadId].mybodytab = Local[0].mybodytab + (maxmybody * ThreadId);
    /* note that every process has its own copy   */
    /* of mybodytab, which was initialized to the */
    /* beginning of the whole array by proc. 0    */
    /* before create                              */
    Local[ThreadId].mycelltab = Local[0].mycelltab + (maxmycell * ThreadId);
    Local[ThreadId].myleaftab = Local[0].myleaftab + (maxmyleaf * ThreadId);
    
    Local[ThreadId].tout = Local[0].tout;
    Local[ThreadId].tnow = Local[0].tnow;
    Local[ThreadId].nstep = Local[0].nstep;
    
    find_my_initial_bodies(bodytab, nbody, ThreadId);
    
    /* main loop */
    while (Local[ThreadId].tnow < tstop + 0.1 * dtime)
    {
	unsigned int istart=0, iend;
	
#ifdef VALIDATE
	Local[ThreadId].totalcost = 0;
	if (ThreadId == 0)
	{
	    Global->totalcost = 0;
	    {long time(long *); (istart) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
	}
#endif VALIDATE
	
	iteration++;
	
	stepsystem(ThreadId);
	
#ifdef VALIDATE
	{GETLOCK(&Global->io_lock);};
	Global->totalcost += Local[ThreadId].totalcost; 
	{FREELOCK(&Global->io_lock);};
	{TREEBAR(&(Global->Barpos), MyNum);};
	if (ThreadId == 0)
	{
	    printf("Total Cost %u\n", Global->totalcost);
	    {long time(long *); (iend) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
	    printf("Iteration %d: %u\n", iteration, iend-istart);
	}
#endif
    }

#ifdef RSIM		   
   endphase(phase);
   if (ThreadId == MASTER)
    	StatReportAll();
#endif

#ifdef TZERO
    threadBarrier();

    if (threadId() == 0)
    {
	dumpStats("End");
    }

    threadBarrier();
#endif TZERO
}



void inputdata();
void pranset(int);
void testdata();
void setbound();

/*
 * STARTRUN: startup hierarchical N-body code.
 */
void
startrun()
{
    string getparam();
    int getiparam();
    short_bool getbparam();
    double getdparam();
    int seed=0;
    
    infile = getparam("in");
    if (*infile != 0) 
    {
	inputdata();
    }
    else 
    {
	nbody = getiparam("nbody");
	if (nbody < 1) 
	{
	    error("startrun: absurd nbody\n");
	}
	seed = getiparam("seed");
    }
    
    outfile = getparam("out");
    dtime = getdparam("dtime");
    dthf = 0.5 * dtime;
    eps = getdparam("eps");
    epssq = eps*eps;
    tol = getdparam("tol");
    tolsq = tol*tol;
    fcells = getdparam("fcells");
    fleaves = getdparam("fleaves");
    tstop = getdparam("tstop");
    dtout = getdparam("dtout");
    Local[0].nstep = 0;
    pranset(seed);
    testdata();
    setbound();
    Local[0].tout = Local[0].tnow + dtout;
    
}

/*
 * TESTDATA: generate Plummer model initial conditions for test runs,
 * scaled to units such that M = -4E = G = 1 (Henon, Hegge, etc).
 * See Aarseth, SJ, Henon, M, & Wielen, R (1974) Astr & Ap, 37, 183.
 */

#define MFRAC  0.999                /* mass cut off at MFRAC of total */

void pickshell(real vec[], real rad);

void
testdata()
{
    real rsc, vsc, sqrt(), xrand(), pow(), r, v, x, y;
    vector cmr, cmv;
    register bodyptr p;
    int rejects = 0;
    int halfnbody, i;
    float offset;
    register bodyptr cp;
    
    headline = "Hack code: Plummer model";
    Local[0].tnow = 0.0;
    bodytab = (bodyptr) shmalloc(nbody * sizeof(body));;
    if (bodytab == NULL) 
    {
	error("testdata: not enuf memory\n");
    }
    rsc = 9 * PI / 16;
    vsc = sqrt(1.0 / rsc);
    
    CLRV(cmr);
    CLRV(cmv);
    
    halfnbody = nbody / 2;
    if (nbody % 2 != 0) halfnbody++;
    for (p = bodytab; p < bodytab+halfnbody; p++) 
    {
	Type(p) = BODY;
	Mass(p) = 1.0 / nbody;
	Cost(p) = 1;
	
	r = 1 / sqrt(pow(xrand(0.0, MFRAC), -2.0/3.0) - 1);
	/*   reject radii greater than 10 */
	while (r > 9.0) 
	{
	    rejects++;
	    r = 1 / sqrt(pow(xrand(0.0, MFRAC), -2.0/3.0) - 1);
	}        
	pickshell(Pos(p), rsc * r);
	ADDV(cmr, cmr, Pos(p));
	do 
	{
	    x = xrand(0.0, 1.0);
	    y = xrand(0.0, 0.1);
	    
	} while (y > x*x * pow(1 - x*x, 3.5));
	
	v = sqrt(2.0) * x / pow(1 + r*r, 0.25);
	pickshell(Vel(p), vsc * v);
	ADDV(cmv, cmv, Vel(p));
    }
    
    offset = 4.0;
    
    for (p = bodytab + halfnbody; p < bodytab+nbody; p++) 
    {
	Type(p) = BODY;
	Mass(p) = 1.0 / nbody;
	Cost(p) = 1;
	
	cp = p - halfnbody;
	for (i = 0; i < NDIM; i++)
	{
	    Pos(p)[i] = Pos(cp)[i] + offset; 
	    ADDV(cmr, cmr, Pos(p));
	    Vel(p)[i] = Vel(cp)[i];
	    ADDV(cmv, cmv, Vel(p));
	}
    }
    
    DIVVS(cmr, cmr, (real) nbody);
    DIVVS(cmv, cmv, (real) nbody);
    
    for (p = bodytab; p < bodytab+nbody; p++) 
    {
	SUBV(Pos(p), Pos(p), cmr);
	SUBV(Vel(p), Vel(p), cmv);
    }
}

/*
 * PICKSHELL: pick a random point on a sphere of specified radius.
 */
void
pickshell(vec, rad)
     real vec[];                     /* coordinate vector chosen */
     real rad;                       /* radius of chosen point */
{
    register int k;
    double rsq, xrand(), sqrt(), rsc;
    
    do 
    {
	for (k = 0; k < NDIM; k++) 
	{
	    vec[k] = xrand(-1.0, 1.0);
	}
	DOTVP(rsq, vec, vec);
    } while (rsq > 1.0);
    
    rsc = rad / sqrt(rsq);
    MULVS(vec, vec, rsc);
}



int 
intpow(i,j)
     int i,j;
{   
    int k;
    int temp = 1;
    
    for (k = 0; k < j; k++)
    {
	temp = temp*i;
    }
    return temp;
}


/*
 * STEPSYSTEM: advance N-body system one time-step.
 */

void maketree(int ThreadId);
void Housekeep(unsigned ThreadId);
void find_my_bodies(nodeptr mycell, int work, int direction, 
		    unsigned ThreadId);

void
stepsystem (ThreadId)
     unsigned int ThreadId;
{
    int i;
    real Cavg;
    bodyptr p,*pp;
    vector dvel, vel1, dpos;
    int intpow();
    unsigned long long trackstart=0, trackend=0;
    unsigned long long partitionstart=0, partitionend=0;
    unsigned long long treebuildstart=0, treebuildend=0;
    unsigned long long forcecalcstart=0, forcecalcend=0;
    
    if (Local[ThreadId].nstep == 2) 
    {
	/* POSSIBLE ENHANCEMENT:  Here is where one might reset the
	   statistics that one is measuring about the parallel execution */
#ifdef RSIM		   
	endphase(phase);
	if (ThreadId == MASTER)
     		StatClearAll();
     	newphase(++phase);
#endif
#ifdef TZERO
    if (threadId() == 0)
    {
	dumpStats("Init");
    }
    threadGlobalBarrier();
#endif TZERO
    }
    
    if ((ThreadId == 0) && (Local[ThreadId].nstep >= 2)) 
    {
        {long time(long *); (trackstart) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    }
    
    if (ThreadId == 0) 
    {
	init_root(ThreadId);
    }
    else 
    {
	Local[ThreadId].mynumcell = 0;
	Local[ThreadId].mynumleaf = 0;
    }
    
    
    /* start at same time */
    {TREEBAR(&(Global->Barstart), MyNum);};
    
    if ((ThreadId == 0) && (Local[ThreadId].nstep >= 2)) 
    {
        {long time(long *); (treebuildstart) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    }
    
    /* load bodies into tree   */
    maketree(ThreadId);
    
    if ((ThreadId == 0) && (Local[ThreadId].nstep >= 2)) 
    {
        {long time(long *); (treebuildend) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
        Global->treebuildtime += treebuildend - treebuildstart;
    }
    
    Housekeep(ThreadId);
    
    Cavg = (real) Cost(Global->G_root) / (real)NTHREAD ;
    Local[ThreadId].workMin = (int) (Cavg * ThreadId);
    Local[ThreadId].workMax = (int) (Cavg * (ThreadId + 1)
				      + (ThreadId == (NTHREAD - 1)));
    
    if ((ThreadId == 0) && (Local[ThreadId].nstep >= 2)) 
    {
        {long time(long *); (partitionstart) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    }
    
    Local[ThreadId].mynbody = 0;
    find_my_bodies((nodeptr)Global->G_root, 0, BRC_FUC, ThreadId );
    
    
    if ((ThreadId == 0) && (Local[ThreadId].nstep >= 2)) 
    {
        {long time(long *); (partitionend) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
        Global->partitiontime += partitionend - partitionstart;
    }
    
    if ((ThreadId == 0) && (Local[ThreadId].nstep >= 2)) 
    {
        {long time(long *); (forcecalcstart) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    }
    
    ComputeForces(ThreadId);
    
    if ((ThreadId == 0) && (Local[ThreadId].nstep >= 2)) 
    {
        {long time(long *); (forcecalcend) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
        Global->forcecalctime += forcecalcend - forcecalcstart;
    }
    
    /* advance my bodies */
    for (pp = Local[ThreadId].mybodytab;
	 pp < Local[ThreadId].mybodytab+Local[ThreadId].mynbody; pp++) 
    {
	p = *pp;
	MULVS(dvel, Acc(p), dthf);              
	ADDV(vel1, Vel(p), dvel);               
	MULVS(dpos, vel1, dtime);               
	ADDV(Pos(p), Pos(p), dpos);             
	ADDV(Vel(p), vel1, dvel);               
        
	for (i = 0; i < NDIM; i++) 
	{
	    if (Pos(p)[i]<Local[ThreadId].min[i]) 
	    {
		Local[ThreadId].min[i]=Pos(p)[i];
	    }
	    if (Pos(p)[i]>Local[ThreadId].max[i]) 
	    {
		Local[ThreadId].max[i]=Pos(p)[i];
	    }
	}
    }
    {GETLOCK(&Global->CountLock);};
    for (i = 0; i < NDIM; i++) 
    {
	if (Global->min[i] > Local[ThreadId].min[i]) 
	{
	    Global->min[i] = Local[ThreadId].min[i];
	}
	if (Global->max[i] < Local[ThreadId].max[i]) 
	{
	    Global->max[i] = Local[ThreadId].max[i];
	}
    }
    {FREELOCK(&Global->CountLock);};
    
    /* bar needed to make sure that every process has computed its min */
    /* and max coordinates, and has accumulated them into the global   */
    /* min and max, before the new dimensions are computed	       */
    {TREEBAR(&(Global->Barpos), MyNum);};
    
    if ((ThreadId == 0) && (Local[ThreadId].nstep >= 2)) 
    {
        {long time(long *); (trackend) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
        Global->tracktime += trackend - trackstart;
    }
    if (ThreadId==0) 
    {
	Global->rsize=0;
	SUBV(Global->max,Global->max,Global->min);
	for (i = 0; i < NDIM; i++) 
	{
	    if (Global->rsize < Global->max[i]) 
	    {
		Global->rsize = Global->max[i];
	    }
	}
	ADDVS(Global->rmin,Global->min,-Global->rsize/100000.0);
	Global->rsize = 1.00002*Global->rsize;
	SETVS(Global->min,1E99);
	SETVS(Global->max,-1E99);
    }
    Local[ThreadId].nstep++;
    Local[ThreadId].tnow = Local[ThreadId].tnow + dtime;
}



void hackgrav(bodyptr p, int ThreadId);

void
ComputeForces (ThreadId)
     unsigned int ThreadId;
{
    bodyptr p,*pp;
    vector acc1, dacc, dvel;
    
    for (pp = Local[ThreadId].mybodytab;
	 pp < Local[ThreadId].mybodytab+Local[ThreadId].mynbody;pp++) 
    {
	p = *pp;
	SETV(acc1, Acc(p));
	Cost(p)=0;
	hackgrav(p,ThreadId);
	Local[ThreadId].myn2bcalc += Local[ThreadId].myn2bterm; 
	Local[ThreadId].mynbccalc += Local[ThreadId].mynbcterm;
	if (!Local[ThreadId].skipself) 
	{       /*   did we miss self-int?  */
	    Local[ThreadId].myselfint++;        /*   count another goofup   */
	}
	if (Local[ThreadId].nstep > 0) 
	{
	    /*   use change in accel to make 2nd order correction to vel    */
	    SUBV(dacc, Acc(p), acc1);
	    MULVS(dvel, dacc, dthf);
	    ADDV(Vel(p), Vel(p), dvel);
	}
    }
}

/* 
 * FIND_MY_INITIAL_BODIES: puts into mybodytab the initial list of bodies 
 * assigned to the thread.  
 */
void
find_my_initial_bodies(btab, nbody, ThreadId)
     bodyptr btab;
     int nbody;
     unsigned int ThreadId;
{
    int intpow();
    int extra,offset=0,i;
    
    Local[ThreadId].mynbody = nbody / NTHREAD;
    extra = nbody % NTHREAD;
    if (ThreadId < extra) 
    {
	Local[ThreadId].mynbody++;    
	offset = Local[ThreadId].mynbody * ThreadId;
    }
    if (ThreadId >= extra) 
    {
	offset = (Local[ThreadId].mynbody+1) * extra + (ThreadId - extra) 
	  * Local[ThreadId].mynbody; 
    }
    for (i=0; i < Local[ThreadId].mynbody; i++) 
    {
	Local[ThreadId].mybodytab[i] = &(btab[offset+i]);
    }
    {TREEBAR(&(Global->Barstart), MyNum);};
}


void
find_my_bodies(mycell, work, direction, ThreadId)
     nodeptr mycell;
     int work;
     int direction;
     unsigned ThreadId;
{
    int i;
    leafptr l;
    nodeptr qptr;
    
    if (Type(mycell) == LEAF) 
    {
	l = (leafptr) mycell;
	for (i = 0; i < l->num_bodies; i++) 
	{
	    if (work >= Local[ThreadId].workMin - .1) 
	    {
		if((Local[ThreadId].mynbody+2) > maxmybody) 
		{
		    error("find_my_bodies: Thread %d needs more than %d bodies; increase fleaves\n",ThreadId, maxmybody); 
		}
		Local[ThreadId].mybodytab[Local[ThreadId].mynbody++] = 
		  Bodyp(l)[i];
	    }
	    work += Cost(Bodyp(l)[i]);
	    if (work >= Local[ThreadId].workMax-.1) 
	    {
		break;
	    }
	}
    }
    else 
    {
	for(i = 0; (i < NSUB) && (work < (Local[ThreadId].workMax - .1)); i++)
	{
	    qptr = Subp(mycell)[Child_Sequence[direction][i]];
	    if (qptr!=NULL) 
	    {
		if ((work+Cost(qptr)) >= (Local[ThreadId].workMin -.1)) 
		{
		    find_my_bodies(qptr,work, Direction_Sequence[direction][i],
				   ThreadId);
		}
		work += Cost(qptr);
	    }
	}
    }
}

/*
 * HOUSEKEEP: reinitialize the different variables (in particular global
 * variables) between each time step.
 */

void
Housekeep(ThreadId)
     unsigned ThreadId;
{
    Local[ThreadId].myn2bcalc = Local[ThreadId].mynbccalc 
      = Local[ThreadId].myselfint = 0;
    SETVS(Local[ThreadId].min,1E99);
    SETVS(Local[ThreadId].max,-1E99);
}

/*
 * SETBOUND: Compute the initial size of the root of the tree; only done
 * before first time step, and only thread 0 does it
 */
void
setbound()
{
    int i;
    real side ;
    bodyptr p;
    
    SETVS(Local[0].min,1E99);
    SETVS(Local[0].max,-1E99);
    side=0;
    
    for (p = bodytab; p < bodytab+nbody; p++) 
    {
	for (i=0; i<NDIM;i++) 
	{
	    if (Pos(p)[i]<Local[0].min[i]) Local[0].min[i]=Pos(p)[i] ;
	    if (Pos(p)[i]>Local[0].max[i])  Local[0].max[i]=Pos(p)[i] ;
	}
    }
    
    SUBV(Local[0].max,Local[0].max,Local[0].min);
    for (i=0; i<NDIM;i++) if (side<Local[0].max[i]) side=Local[0].max[i];
    ADDVS(Global->rmin,Local[0].min,-side/100000.0);
    Global->rsize = 1.00002*side;
    SETVS(Global->max,-1E99);
    SETVS(Global->min,1E99);
}

void
Help () 
{
    printf("The only on-line paramter is the number of threads per node\n");
    printf("The rest of the parameters should be placed in a file\n");
    printf("There are a total of twelve parameters, and all of them have default values.\n");
    printf("\n");
    printf("1) infile (char*) : The name of an input file that contains particle data.  \n");
    printf("    The format of the file is:\n");
    printf("\ta) An int representing the number of particles in the distribution\n");
    printf("\tb) An int representing the dimensionality of the problem (3-D)\n");
    printf("\tc) A double representing the current time of the simulation\n");
    printf("\td) Doubles representing the masses of all the particles\n");
    printf("\te) A vector (length equal to the dimensionality) of doubles\n");
    printf("\t   representing the positions of all the particles\n");
    printf("\tf) A vector (length equal to the dimensionality) of doubles\n");
    printf("\t   representing the velocities of all the particles\n");
    printf("\n");
    printf("    Each of these numbers can be separated by any amount of whitespace.\n");
    printf("\n");
    printf("2) nbody (int) : If no input file is specified (the first line is blank), this\n");
    printf("    number specifies the number of particles to generate under a plummer model.\n");
    printf("    Default is 16384.\n");
    printf("\n");
    printf("3) seed (int) : The seed used by the random number generator.\n");
    printf("    Default is 123.\n");
    printf("\n");
    printf("4) outfile (char*) : The name of the file that snapshots will be printed to. \n");
    printf("    This feature has been disabled in the SPLASH release.\n");
    printf("    Default is NULL.\n");
    printf("\n");
    printf("5) dtime (double) : The integration time-step.\n");
    printf("    Default is 0.025.\n");
    printf("\n");
    printf("6) eps (double) : The usual potential softening\n");
    printf("    Default is 0.05.\n");
    printf("\n");
    printf("7) tol (double) : The cell subdivision tolerance.\n");
    printf("    Default is 1.0.\n");
    printf("\n");
    printf("8) fcells (double) : The total number of cells created is equal to \n");
    printf("    fcells * number of leaves.\n");
    printf("    Default is 2.0.\n");
    printf("\n");
    printf("9) fleaves (double) : The total number of leaves created is equal to  \n");
    printf("    fleaves * nbody.\n");
    printf("    Default is 0.5.\n");
    printf("\n");
    printf("10) tstop (double) : The time to stop integration.\n");
    printf("    Default is 0.075.\n");
    printf("\n");
    printf("11) dtout (double) : The data-output interval.\n");
    printf("    Default is 0.25.\n");
    printf("\n");
}
