#line 1 "-"
#line 1 "/home/lagavulin/a/gniady/rsim-1.0/apps/utils/lib/rsim_macros"

#line 120

#line 1 "src/code.H"
/*************************************************************************/
/*                                                                       */
/*  Copyright (c) 1994 Stanford University                               */
/*                                                                       */
/*  All rights reserved.                                                 */
/*                                                                       */
/*  Permission is given to use, copy, and modify this software for any   */
/*  non-commercial purpose as long as this copyright notice is not       */
/*  removed.  All other uses, including redistribution in whole or in    */
/*  part, are forbidden without prior written permission.                */
/*                                                                       */
/*  This software is provided with absolutely no warranty and no         */
/*  support.                                                             */
/*                                                                       */
/*************************************************************************/

/*
 * CODE.H:  various global things for CODE.C.
 */

#ifndef _CODE_H_
#define _CODE_H_

#include "defs.h"

#define PAD_SIZE (PAGE_SIZE / (sizeof(int)))

#define NUMBER_OF_PROCESSORS    NTHREAD
#define MyNum                   ThreadId
PUBLIC int NTHREAD; 		/* number of threads */

/* Defined by the input file */
PUBLIC string headline; 	/* message describing calculation */
PUBLIC string infile; 		/* file name for snapshot input */
PUBLIC string outfile; 		/* file name for snapshot output */
PUBLIC real dtime; 		/* timestep for leapfrog integrator */
PUBLIC real dtout; 		/* time between data outputs */
PUBLIC real tstop; 		/* time to stop calculation */
PUBLIC int nbody; 		/* number of bodies in system */
PUBLIC real fcells; 		/* ratio of cells/leaves allocated */
PUBLIC real fleaves; 		/* ratio of leaves/bodies allocated */
PUBLIC real tol; 		/* accuracy parameter: 0.0 => exact */
PUBLIC real tolsq; 		/* square of previous */
PUBLIC real eps; 		/* potential softening parameter */
PUBLIC real epssq; 		/* square of previous */
PUBLIC real dthf; 		/* half time step */
PUBLIC int NTHREADPERNODE; 	/* Number of Threads per node */

PUBLIC int maxcell;		/* max number of cells allocated */
PUBLIC int maxleaf;		/* max number of leaves allocated */
PUBLIC int maxmybody;		/* max no. of bodies allocated per thread */
PUBLIC int maxmycell;		/* max num. of cells to be allocated */
PUBLIC int maxmyleaf;		/* max num. of leaves to be allocated */
PUBLIC bodyptr bodytab; 	/* array size is exactly nbody bodies */

PUBLIC struct CellLockType {
    volatile int CL[MAXLOCK];        /* locks on the cells*/
} *CellLock;
    
struct GlobalMemory  {	/* all this info is for the whole system */
    int n2bcalc;       /* total number of body/cell interactions  */
    int nbccalc;       /* total number of body/body interactions  */
    int selfint;       /* number of self interactions             */
    real mtot;         /* total mass of N-body system             */
    real etot[3];      /* binding, kinetic, potential energy      */
    matrix keten;      /* kinetic energy tensor                   */
    matrix peten;      /* potential energy tensor                 */
    vector cmphase[2]; /* center of mass coordinates and velocity */
    vector amvec;      /* angular momentum vector                 */
    cellptr G_root;    /* root of the whole tree                  */
    vector rmin;       /* lower-left corner of coordinate box     */
    vector min;        /* temporary lower-left corner of the box  */
    vector max;        /* temporary upper right corner of the box */
    real rsize;        /* side-length of integer coordinate box   */
    TreeBar Barstart;   /* barrier at the beginning of stepsystem  */
    TreeBar Barover;    /* SHAN */
    TreeBar Bartree;    /* barrier after loading the tree          */
    TreeBar Barcom;     /* barrier after computing the c. of m.    */
    TreeBar Barload;    /*   */
    TreeBar Baraccel;   /* barrier after accel and before output   */
    TreeBar Barpos;     /* barrier after computing the new pos     */
    volatile int (CountLock); /* Lock on the shared variables            */
    volatile int (NcellLock); /* Lock on the counter of array of cells for loadtree */
    volatile int (NleafLock); /* Lock on the counter of array of leaves for loadtree*/
    volatile int (io_lock);
    unsigned long long createstart,createend,computestart,computeend;
    unsigned long long trackstart, trackend, tracktime;
    unsigned long long partitionstart, partitionend, partitiontime;
    unsigned long long treebuildstart, treebuildend, treebuildtime;
    unsigned long long forcecalcstart, forcecalcend, forcecalctime;
    unsigned long long current_id;
    volatile int k; /*for memory allocation in code.C */

#ifdef VALIDATE
    int totalcost;
#endif

};
PUBLIC struct GlobalMemory *Global;

/* This structure is needed because under the sproc model there is no
 * per thread private address space. 
 */
struct local_memory 
{
   /* Use padding so that each thread's variables are on their own page */
   int pad_begin[PAD_SIZE];

   real tnow;        	/* current value of simulation time */
   real tout;         	/* time next output is due */
   int nstep;      	/* number of integration steps so far */

   int workMin, workMax;/* interval of cost to be treated by a thread */

   vector min, max; 	/* min and max of coordinates for each thread. */

   int mynumcell; 	/* num. of cells used for this proc in ctab */
   int mynumleaf; 	/* num. of leaves used for this proc in ctab */
   int mynbody;   	/* num bodies allocated to the thread */
   bodyptr* mybodytab;	/* array of bodies allocated / thread */
   int myncell; 	/* num cells allocated to the thread */
   cellptr* mycelltab;	/* array of cellptrs allocated to the thread */
   int mynleaf; 	/* number of leaves allocated to the thread */
   leafptr* myleaftab; 	/* array of leafptrs allocated to the thread */
   cellptr ctab;	/* array of cells used for the tree. */
   leafptr ltab;	/* array of cells used for the tree. */

   int myn2bcalc; 	/* body-body force calculations for each thread */
   int mynbccalc; 	/* body-cell force calculations for each thread */
   int myselfint; 	/* count self-interactions for each thread */
   int myn2bterm; 	/* count body-body terms for a body */
   int mynbcterm; 	/* count body-cell terms for a body */
   short_bool skipself; 	/* true if self-interaction skipped OK */
   bodyptr pskip;       /* body to skip in force evaluation */
   vector pos0;         /* point at which to evaluate field */
   real phi0;           /* computed potential at pos0 */
   vector acc0;         /* computed acceleration at pos0 */
   vector dr;  		/* data to be shared */
   real drsq;      	/* between gravsub and subdivp */
   nodeptr pmem;	/* remember particle data */

   nodeptr Current_Root;
   int Root_Coords[NDIM];

   real mymtot;      	/* total mass of N-body system */
   real myetot[3];   	/* binding, kinetic, potential energy */
   matrix myketen;   	/* kinetic energy tensor */
   matrix mypeten;   	/* potential energy tensor */
   vector mycmphase[2];	/* center of mass coordinates */
   vector myamvec;   	/* angular momentum vector */

#ifdef VALIDATE
    int totalcost;
#endif

   int pad_end[PAD_SIZE];
};
PUBLIC struct local_memory Local[MAX_THREAD];

#endif
