#line 1 "-"
#line 1 "/home/lagavulin/a/gniady/rsim-1.0/apps/utils/lib/rsim_macros"

#line 120

#line 1 "src/code_io.C"
/*************************************************************************/
/*                                                                       */
/*  Copyright (c) 1994 Stanford University                               */
/*                                                                       */
/*  All rights reserved.                                                 */
/*                                                                       */
/*  Permission is given to use, copy, and modify this software for any   */
/*  non-commercial purpose as long as this copyright notice is not       */
/*  removed.  All other uses, including redistribution in whole or in    */
/*  part, are forbidden without prior written permission.                */
/*                                                                       */
/*  This software is provided with absolutely no warranty and no         */
/*  support.                                                             */
/*                                                                       */
/*************************************************************************/

/*
 * CODE_IO.C: 
 */
#ifdef WWT
{;} 
#else
#include <rsim_apps.h>
#endif
#define PUBLIC extern

#include "code.h"

void in_int (), in_real (), in_vector ();
void out_int (), out_real (), out_vector ();
void diagnostics (unsigned int ThreadId);

/*
 * INPUTDATA: read initial conditions from input file.
 */
void    
inputdata ()
{
    stream instr;
    static char headbuf[128];
    int ndim;
    real tnow;
    bodyptr p;
    int i;
    
    fprintf(stderr,"reading input file : %s\n",infile);
    fflush(stderr);
    instr = fopen(infile, "r");
    if (instr == NULL)
    {
	error("inputdata: cannot find file %s\n", infile);
    }
    sprintf(headbuf, "Hack code: input file %s\n", infile);
    headline = headbuf;
    in_int(instr, &nbody);
    if (nbody < 1)
    {
	error("inputdata: nbody = %d is absurd\n", nbody);
    }
    in_int(instr, &ndim);
    if (ndim != NDIM)
    {
	error("inputdata: NDIM = %d ndim = %d is absurd\n", NDIM,ndim);
    }
    in_real(instr, &tnow);
    for (i = 0; i < MAX_THREAD; i++)
    {
	Local[i].tnow = tnow;
    }
    bodytab = (bodyptr) shmalloc(nbody * sizeof(body));;
    if (bodytab == NULL)
    {
	error("inputdata: not enuf memory\n");
    }
    for (p = bodytab; p < bodytab+nbody; p++) 
    {
	Type(p) = BODY;
	Cost(p) = 1;
	Phi(p) = 0.0;
	CLRV(Acc(p));
    } 
    for (p = bodytab; p < bodytab+nbody; p++)
    {
	in_real(instr, &Mass(p));
    }
    for (p = bodytab; p < bodytab+nbody; p++)
    {
	in_vector(instr, Pos(p));
    }
    for (p = bodytab; p < bodytab+nbody; p++)
    {
	in_vector(instr, Vel(p));
    }
    fclose(instr);
}

/*
 * INITOUTPUT: initialize output routines.
 */

void
initoutput()
{
    printf("running %d thread(s)\n", NTHREAD);
    printf("\n\t\t%s\n\n", headline);
    printf("%10s%10s%10s%10s%10s%10s%10s%10s\n",
	   "nbody", "dtime", "eps", "tol", "dtout", "tstop","fcells",
	   "NTHREAD");
    printf("%10d%10.5f%10.4f%10.2f%10.3f%10.3f%10.2f%10d\n\n",
	   nbody, dtime, eps, tol, dtout, tstop, fcells, NTHREAD);
}

/*
 * STOPOUTPUT: finish up after a run.
 */


/*
 * OUTPUT: compute diagnostics and output data.
 */

void
output (ThreadId)
     unsigned int ThreadId;
{
    int nttot, nbavg, ncavg,k;
    double cputime();
    vector tempv1,tempv2;
    
    if ((Local[ThreadId].tout - 0.01 * dtime) <= Local[ThreadId].tnow) 
    {
	Local[ThreadId].tout += dtout;
    }
    
    diagnostics(ThreadId);
    
    if (Local[ThreadId].mymtot!=0) 
    {
	{GETLOCK(&Global->CountLock);};
	Global->n2bcalc += Local[ThreadId].myn2bcalc;
	Global->nbccalc += Local[ThreadId].mynbccalc;
	Global->selfint += Local[ThreadId].myselfint;
	ADDM(Global->keten, Global-> keten, Local[ThreadId].myketen);
	ADDM(Global->peten, Global-> peten, Local[ThreadId].mypeten);
	for (k=0;k<3;k++) Global->etot[k] +=  Local[ThreadId].myetot[k];
	ADDV(Global->amvec, Global-> amvec, Local[ThreadId].myamvec);
        
	MULVS(tempv1, Global->cmphase[0],Global->mtot);
	MULVS(tempv2, Local[ThreadId].mycmphase[0], Local[ThreadId].mymtot);
	ADDV(tempv1, tempv1, tempv2);
	DIVVS(Global->cmphase[0], tempv1, 
	      Global->mtot+Local[ThreadId].mymtot); 
        
	MULVS(tempv1, Global->cmphase[1],Global->mtot);
	MULVS(tempv2, Local[ThreadId].mycmphase[1], Local[ThreadId].mymtot);
	ADDV(tempv1, tempv1, tempv2);
	DIVVS(Global->cmphase[1], tempv1, 
	      Global->mtot+Local[ThreadId].mymtot);
	Global->mtot +=Local[ThreadId].mymtot;
	{FREELOCK(&Global->CountLock);};
    }
    
    {TREEBAR(&(Global->Baraccel), MyNum);};
    
    if (ThreadId==0) 
    {
	nttot = Global->n2bcalc + Global->nbccalc;
	nbavg = (int) ((real) Global->n2bcalc / (real) nbody);
	ncavg = (int) ((real) Global->nbccalc / (real) nbody);
    }
}



/*
 * DIAGNOSTICS: compute set of dynamical diagnostics.
 */

void
diagnostics (ThreadId)
     unsigned int ThreadId;
{
    register bodyptr p,*pp;
    real velsq;
    vector tmpv;
    matrix tmpt;
    
    Local[ThreadId].mymtot = 0.0;
    Local[ThreadId].myetot[1] = Local[ThreadId].myetot[2] = 0.0;
    CLRM(Local[ThreadId].myketen);
    CLRM(Local[ThreadId].mypeten);
    CLRV(Local[ThreadId].mycmphase[0]);
    CLRV(Local[ThreadId].mycmphase[1]);
    CLRV(Local[ThreadId].myamvec);
    for (pp = Local[ThreadId].mybodytab+Local[ThreadId].mynbody -1; 
	 pp >= Local[ThreadId].mybodytab; pp--) 
    { 
	p= *pp;
	Local[ThreadId].mymtot += Mass(p);
	DOTVP(velsq, Vel(p), Vel(p));
	Local[ThreadId].myetot[1] += 0.5 * Mass(p) * velsq;
	Local[ThreadId].myetot[2] += 0.5 * Mass(p) * Phi(p);
	MULVS(tmpv, Vel(p), 0.5 * Mass(p));
	OUTVP(tmpt, tmpv, Vel(p));
	ADDM(Local[ThreadId].myketen, Local[ThreadId].myketen, tmpt);
	MULVS(tmpv, Pos(p), Mass(p));
	OUTVP(tmpt, tmpv, Acc(p));
	ADDM(Local[ThreadId].mypeten, Local[ThreadId].mypeten, tmpt);
	MULVS(tmpv, Pos(p), Mass(p));
	ADDV(Local[ThreadId].mycmphase[0], Local[ThreadId].mycmphase[0], 
	     tmpv);
	MULVS(tmpv, Vel(p), Mass(p));
	ADDV(Local[ThreadId].mycmphase[1], Local[ThreadId].mycmphase[1], 
	     tmpv);
	CROSSVP(tmpv, Pos(p), Vel(p));
	MULVS(tmpv, tmpv, Mass(p));
	ADDV(Local[ThreadId].myamvec, Local[ThreadId].myamvec, tmpv);
    }
    Local[ThreadId].myetot[0] = Local[ThreadId].myetot[1] 
      + Local[ThreadId].myetot[2];
    if (Local[ThreadId].mymtot!=0)
    {
	DIVVS(Local[ThreadId].mycmphase[0], Local[ThreadId].mycmphase[0], 
	      Local[ThreadId].mymtot);
	DIVVS(Local[ThreadId].mycmphase[1], Local[ThreadId].mycmphase[1], 
	      Local[ThreadId].mymtot);
    }
}



/*
 * Low-level input and output operations.
 */

void in_int(str, iptr)
     stream str;
     int *iptr;
{
    if (fscanf(str, "%d", iptr) != 1)
    {
	error("in_int: input conversion error\n");
    }
}

void in_real(str, rptr)
     stream str;
     real *rptr;
{
    double tmp;
    
    if (fscanf(str, "%lf", &tmp) != 1)
    {
	error("in_real: input conversion error\n");
    }
    *rptr = tmp;
}

void in_vector(str, vec)
     stream str;
     vector vec;
{
    double tmpx, tmpy, tmpz;
    
    if (fscanf(str, "%lf%lf%lf", &tmpx, &tmpy, &tmpz) != 3)
    {
	error("in_vector: input conversion error\n");
    }
    vec[0] = tmpx;    vec[1] = tmpy;    vec[2] = tmpz;
}

void out_int(str, ival)
     stream str;
     int ival;
{
    fprintf(str, "  %d\n", ival);
    fflush(stderr);
}

void out_real(str, rval)
     stream str;
     real rval;
{
    fprintf(str, " %21.14E\n", rval);
    fflush(stderr);
}

void out_vector(str, vec)
     stream str;
     vector vec;
{
    fprintf(str, " %21.14E %21.14E", vec[0], vec[1]);
    fprintf(str, " %21.14E\n",vec[2]);
    fflush(stderr);
}
