/*************************************************************************/
/*                                                                       */
/*  Copyright (c) 1994 Stanford University                               */
/*                                                                       */
/*  All rights reserved.                                                 */
/*                                                                       */
/*  Permission is given to use, copy, and modify this software for any   */
/*  non-commercial purpose as long as this copyright notice is not       */
/*  removed.  All other uses, including redistribution in whole or in    */
/*  part, are forbidden without prior written permission.                */
/*                                                                       */
/*  This software is provided with absolutely no warranty and no         */
/*  support.                                                             */
/*                                                                       */
/*************************************************************************/

/*
 * GRAV.C: 
 */

#ifdef WWT
ENV
#else
EXTERN_ENV
#endif

#define PUBLIC extern

#include "code.h"

void hackwalk(proced sub, unsigned ThreadId);
void gravsub();

/*
 * HACKGRAV: evaluate grav field at a given particle.
 */
void  
hackgrav(p,ThreadId)
     bodyptr p;
     unsigned ThreadId;
     
{
    Local[ThreadId].pskip = p;
    SETV(Local[ThreadId].pos0, Pos(p));
    Local[ThreadId].phi0 = 0.0;
    CLRV(Local[ThreadId].acc0);
    Local[ThreadId].myn2bterm = 0;
    Local[ThreadId].mynbcterm = 0;
    Local[ThreadId].skipself = FALSE;
    hackwalk(gravsub, ThreadId);
    Phi(p) = Local[ThreadId].phi0;
    SETV(Acc(p), Local[ThreadId].acc0);
#ifdef QUADPOLE
    Cost(p) = Local[ThreadId].myn2bterm + NDIM * Local[ThreadId].mynbcterm;
#else
    Cost(p) = Local[ThreadId].myn2bterm + Local[ThreadId].mynbcterm;
#endif
    
#ifdef VALIDATE
    Local[ThreadId].totalcost += Local[ThreadId].myn2bterm + 
      Local[ThreadId].mynbcterm;
#endif
}



/*
 * GRAVSUB: compute a single body-body or body-cell interaction.
 */
void
gravsub(p, ThreadId, level)
     register nodeptr p;               /* body or cell to interact with     */
     unsigned ThreadId;
     int level;
{
    double sqrt();
    real drabs, phii, mor3;
    vector ai;
#ifdef QUADPOLE
    vector quaddr;
    real dr5inv, phiquad, drquaddr;
#endif QUADPOLE
    
    if (p != Local[ThreadId].pmem) 
    {
        SUBV(Local[ThreadId].dr, Pos(p), Local[ThreadId].pos0);
        DOTVP(Local[ThreadId].drsq, Local[ThreadId].dr, Local[ThreadId].dr);
    }
    
    Local[ThreadId].drsq += epssq;
    drabs = sqrt((double) Local[ThreadId].drsq);
    phii = Mass(p) / drabs;
    Local[ThreadId].phi0 -= phii;
    mor3 = phii / Local[ThreadId].drsq;
    MULVS(ai, Local[ThreadId].dr, mor3);
    ADDV(Local[ThreadId].acc0, Local[ThreadId].acc0, ai); 
    if(Type(p) != BODY) 
    {                  /* a body-cell/leaf interaction? */
	Local[ThreadId].mynbcterm++;
#ifdef QUADPOLE
	dr5inv = 1.0/(Local[ThreadId].drsq * Local[ThreadId].drsq * drabs);
	MULMV(quaddr, Quad(p), Local[ThreadId].dr);
	DOTVP(drquaddr, Local[ThreadId].dr, quaddr);
	phiquad = -0.5 * dr5inv * drquaddr;
	Local[ThreadId].phi0 += phiquad;
	phiquad = 5.0 * phiquad / Local[ThreadId].drsq;
	MULVS(ai, Local[ThreadId].dr, phiquad);
	SUBV(Local[ThreadId].acc0, Local[ThreadId].acc0, ai);
	MULVS(quaddr, quaddr, dr5inv);   
	SUBV(Local[ThreadId].acc0, Local[ThreadId].acc0, quaddr);
#endif
    }
    else 
    {                                      /* a body-body interaction  */
	Local[ThreadId].myn2bterm++;
    }
}

/*
 * HACKWALK: walk the tree opening cells too close to a given point.
 */

void walksub(nodeptr n, real dsq, unsigned ThreadId);

void
hackwalk(sub, ThreadId)
     proced sub;                                /* routine to do calculation */
     unsigned ThreadId;
{
    walksub((nodeptr)Global->G_root, Global->rsize * Global->rsize, ThreadId);
}

/*
 * WALKSUB: recursive routine to do hackwalk operation.
 */

void
walksub(n, dsq, ThreadId)
     nodeptr n;                        /* pointer into body-tree    */
     real dsq;                         /* size of box squared       */
     unsigned ThreadId;
{
    short_bool subdivp();
    nodeptr* nn;
    leafptr l;
    bodyptr p;
    int i;
    
    if (subdivp(n, dsq, ThreadId)) 
    {
	if (Type(n) == CELL) 
	{
	    for (nn = Subp(n); nn < Subp(n) + NSUB; nn++) 
	    {
		if (*nn != NULL) 
		{
		    walksub(*nn, dsq / 4.0, ThreadId);
		}
	    }
	}
	else 
	{
	    l = (leafptr) n;
	    for (i = 0; i < l->num_bodies; i++) 
	    {
		p = Bodyp(l)[i];
		if (p != Local[ThreadId].pskip) 
		{
		    gravsub(p, ThreadId);
		}
		else 
		{
		    Local[ThreadId].skipself = TRUE;
		}
	    }
	}
    }
    else 
    {
	gravsub(n, ThreadId);
    }
}

/*
 * SUBDIVP: decide if a node should be opened.
 * Side effects: sets  pmem,dr, and drsq.
 */

short_bool subdivp(p, dsq, ThreadId)
     register nodeptr p;                      /* body/cell to be tested    */
     real dsq;                                /* size of cell squared      */
     unsigned ThreadId;
{
    SUBV(Local[ThreadId].dr, Pos(p), Local[ThreadId].pos0);
    DOTVP(Local[ThreadId].drsq, Local[ThreadId].dr, Local[ThreadId].dr);
    Local[ThreadId].pmem = p;
    return (tolsq * Local[ThreadId].drsq < dsq);
}
