#line 1 "-"
#line 1 "/home/lagavulin/a/gniady/rsim-1.0/apps/utils/lib/rsim_macros"

#line 120

#line 1 "src/em3d.C"
/*
 * Changes copyright (c) 1992-94 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must  the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 *
 * This source file is part of the Wisconsin Wind Tunnel,
 * originally written by Steven K. Reinhardt and David A. Wood,
 * and subsequently maintained and enhanced by Steven K. Reinhardt,
 * Babak Falsafi, and Alvin R. Lebeck.
 *
 * The TEMPEST version of EM3D was originally written by Steve Reinhardt
 * and subsequently rewritten by Babak Falsafi.
 *
 */

/*
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO {;} SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: 			Steve Lumetta
 * Version:			23
 * Creation Date:	Sat Jan 23 12:45:18 1993
 * Filename:		em3d.sc
 * History:
 *	SL	1	Sat Jan 23 12:48:43 1993
 *		Prepared for release as example.
 */

#include <malloc.h>

#ifdef WWT
{;} 
#else
#include <rsim_apps.h>
TreeBar final_sync;
#line 64
#endif

#define _EM3D_MAIN
#include <stddef.h>
#include <stdio.h>

#include "em3d.h"
#include "prot.h"

#ifdef BLZ
#include "syscalls.h"
#include "cyklos/interface.h"
#include "tppi_lib_vm.h"
#include "tppi_thread.h"
#include "shm_malloc.h"
#endif

#ifdef BLZACC_hwstache
#include "hwstache.h"
#endif

#ifdef RSIM
	int phase = 0;
#endif

void checksum(int my_thread, node_t *node_vec, int n_nodes)
{
    node_t *cur_node;
    int n;

    for (n = 0, cur_node = node_vec; n < n_nodes; ++n, ++cur_node)
    {
	shmem->check_sum[my_thread] += node_val(cur_node);
    }
}

void all_check_sum(int my_thread)
{
    shmem->check_sum[my_thread] = 0.0;
    
    checksum(my_thread, h_nodes, n_nodes);
    checksum(my_thread, e_nodes, n_nodes);
	
    {TREEBAR(&(shmem->check_bar), MyNum);};
    
    if (my_thread == 0)
    {
	int i;
	
	for (i = 1; i < num_threads; i++)
	{
	    shmem->check_sum[0] += shmem->check_sum[i];
	}
    }

    {TREEBAR(&(shmem->check_bar), MyNum);};
}

void print_graph(int my_thread)
{
    int p, i, j;
    node_t *cp;

    for (p = 0; p < num_threads; ++p)
    {
	if (my_thread == p)
	{
	    for (i = 0; i < n_nodes; i++)
	    {
		cp = &h_nodes[i];

		printf("%d: h node %d (%f) :\n", p, i, node_val(cp));

		for (j = 0; j < cp->from_nodes.length; j++)
		{
		    printf("\t value %f coeff %f\n",
			   *cp->from_nodes.values[j],
			   cp->from_nodes.coeffs[j]);
		}
		printf("\n");
	    }
	}
	{TREEBAR(&(shmem->print_bar), MyNum);};
    }

    for (p = 0; p < num_threads; ++p)
    {
	if (my_thread == p)
	{
	    for (i = 0; i < n_nodes; i++)
	    {
		cp = &e_nodes[i];

		printf("%d: e node %d (%f) :\n", p, i, node_val(cp));

		for (j = 0; j < cp->from_nodes.length; j++)
		{
		    printf("\t value %f coeff %f\n",
			   *cp->from_nodes.values[j],
			   cp->from_nodes.coeffs[j]);
		}
		printf("\n");
	    }
	}
	{TREEBAR(&(shmem->print_bar), MyNum);};
    }
}

/* 
 * all_compute_e performs a single time step on the e nodes--each node is
 * adjusted with each of its from links.
 */
void all_compute(node_t *node_vec, int n_nodes)
{
    node_t *cur_node;
    int n, i;

    for (n = 0, cur_node = node_vec; n < n_nodes; ++n, ++cur_node) 
    {
	register double cur_value;
	
	cur_value = node_val(cur_node);
	
	for (i = 0; i < cur_node->from_nodes.length-1;) 
	{
	    register double *p1, *p2;
	    register double d1, d2, d3, d4, d5, d6;
	    
	    p1 = cur_node->from_nodes.values[i];
	    d1 = cur_node->from_nodes.coeffs[i];
	    d2 = *p1;
	    p2 = cur_node->from_nodes.values[i+1];
	    d5 = d1*d2;
	    d3 = *p2;
	    d4 = cur_node->from_nodes.coeffs[i+1];
	    cur_value -= d5;
	    d6 = d3*d4;
	    i += 2;
	    cur_value -= d6;
	}
	
	if (i == cur_node->from_nodes.length-1) 
	{
	    cur_value -= (*cur_node->from_nodes.values[i])*
	      cur_node->from_nodes.coeffs[i];
	}
	
#ifdef SHADOW
	shadow_values[n] = cur_value;
#else
	node_val(cur_node) = cur_value;
#endif
    }

#ifdef SHADOW
    for (n = 0; n < n_nodes; ++n) 
    {
	node_val(&node_vec[n]) = shadow_values[n];
    }
#endif
}

void thread_main()
{
    void srandom(int);
    int my_thread;
    int t = 0;
    int edges = (2*n_nodes)*d_nodes;
    unsigned long long int 
      start_time, 
      init_time, 
      first_iter_time, 
      compute_time;

#ifdef RSIM
    my_thread = getpid();
#else
    threadGlobalBarrier();
    my_thread = threadGlobalId();
#endif

    if (my_thread == 0)
    {
	printf("Running %d thread(s) and %d node(s)\n", 
               num_threads, n_nodes);
    }

    /* The stuff below does not REALLY need to be allocated for all
     * threads on every node. I don't have time to do it right.
     */

    srandom(my_thread * 17);

    /* Seed the random number generator. */
    initProt(my_thread);

    /* Initialize the graph. */

    if (my_thread == 0)
    {
	printf("Building graph...\n");
    }

    {TREEBAR(&(shmem->main_bar), MyNum);};

    {long time(long *); (start_time) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;

    all_initialize_graph(my_thread);

    {long time(long *); (init_time) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;

    init_time -= start_time;

    {TREEBAR(&(shmem->main_bar), MyNum);};

    /* Compute! */
    if (my_thread == 0)
    {
	printf("Computing First Iter...\n");
    }

#ifdef BLZACC_hwstache
    if (threadId() == 0)
    {
	mark_pages_for_migration();
    }
    threadGlobalBarrier();
#endif

    {long time(long *); (start_time) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;

    all_compute(e_nodes, n_nodes);
    end_first_e_iter(my_thread);
    all_compute(h_nodes, n_nodes);
    end_first_h_iter(my_thread);

    {long time(long *); (first_iter_time) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;

    first_iter_time -= start_time;

    if (my_thread == 0)
    {
	printf("Computing Rest...\n");
    }

    {TREEBAR(&(shmem->main_bar), MyNum);};

#ifdef TZERO
    if (threadId() == 0)
    {
	dumpStats("Init");
    }
    threadGlobalBarrier();
#endif TZERO

#ifdef RSIM		   
   endphase(phase);
   if (MyNum == 0)
	StatClearAll();
   newphase(++phase);
#endif

    {long time(long *); (start_time) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;

    for (t = 1; t < time_steps; t++) 
    {
        all_compute(e_nodes, n_nodes);
	end_iter(my_thread, e_nodes, n_nodes);
	all_compute(h_nodes, n_nodes);
	end_iter(my_thread, h_nodes, n_nodes);
    }

    {long time(long *); (compute_time) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;

    compute_time -= start_time;

    end_all_iters();

#ifdef TZERO
    threadBarrier();
    if (threadId() == 0)
    {
	dumpStats("End");
    }
    threadBarrier();
#endif TZERO

#ifdef RSIM		   
   endphase(phase);
   if (MyNum == 0)
   {
   	StatReportAll();
   }
#endif

    if (do_checksum)
    {
	all_check_sum(my_thread);
	if (my_thread == 0)
	{
	    printf("Checksum %f\n", shmem->check_sum[0]);
	}
    }

    if (my_thread == 0) 
    {
	printf("%7s %15s %8s %15s\n",
	       "Edges","Init","First","Compute");
	printf("%7d %15.0f %8.0f %15.0f\n",
		     edges, 
		     (double)init_time,
		     (double)first_iter_time,
		     (double)compute_time);
	printf("Initialization = %1.0f\n", 
	       (double)(init_time + first_iter_time));
	printf("Computation = %1.0f\n", 
	       (double)(compute_time));
	fflush(stdout);

#ifdef BLZ
	printf("Communication = send %1.0f recv %1.0f\n", 
	       (double)(threadData[0].update_time),
	       (double)(threadData[0].receive_time));
#endif
    }
    {TREEBAR(&(shmem->main_bar), MyNum);};
}

#if defined(BLZ) || defined(TZERO)
void real_main
#else
void main
#endif
(int argc, char **argv)
{
    int arg = 0;

    /* Now change argc & argv so we don't have to modify other arg code */
    --argc;
    ++argv;

#ifdef RSIM
MEMSYS_OFF;
#endif    
    
    if (argc == 0)
    {
	printf("Must at least specify proc num!\n");
	exit(-1);
    }

    if (argc > 0)  /* Number of threads */
    {
	sscanf(argv[arg++], "%d", &num_threads);
    }

    if (argc > 1)  /* Number of nodes */
    {
	sscanf(argv[arg++], "%d", &n_nodes);
    }

    if (argc > 2)  /* Degree of nodes */
    {
	sscanf(argv[arg++], "%d", &d_nodes);
    }

    if (argc > 3)  /* Percent of remote links */
    {
	if (sscanf(argv[arg++], "%d", &local_p))
	{
	    local_p = 100 - local_p;
	}
    }

    if (argc > 4)  /* Distribution span */
    {
	sscanf(argv[arg++], "%d", &dist_span);
    }

    if (argc > 5)  /* time steps */
    {
	sscanf(argv[arg++], "%d", &time_steps);
    }

    if (argc > 6)  /* checksum */
    {
	sscanf(argv[arg++], "%d", &do_checksum);
    }

#ifdef BLZ
    INITENV(0);
#else
    {{TreeBarInit(&(final_sync), NUMBER_OF_PROCESSORS);};};
#endif

    n_nodes /= num_threads;
    
    printf("%d nodes/thread, degree %d, ",
	   n_nodes, d_nodes);
    printf("%d%% remote, span %d, %d steps\n", 100-local_p, dist_span, 
	   time_steps);

    shmem = (shmem_t *)shmalloc(sizeof(shmem_t));;
    shmem->h_table = (node_t **)shmalloc(sizeof(node_t *) * num_threads);;
    shmem->e_table = (node_t **)shmalloc(sizeof(node_t *) * num_threads);;
    shmem->check_sum = (double *)shmalloc(sizeof(double) * num_threads);;

    {TreeBarInit(&(shmem->check_bar), NUMBER_OF_PROCESSORS);}
    {TreeBarInit(&(shmem->print_bar), NUMBER_OF_PROCESSORS);}
    {TreeBarInit(&(shmem->main_bar), NUMBER_OF_PROCESSORS);}

#ifdef NO_FORK_SEMANTICS
    set_fork_semantics(&shmem, sizeof(shmem));
    set_fork_semantics(&n_nodes, sizeof(n_nodes));
    set_fork_semantics(&d_nodes, sizeof(d_nodes));
    set_fork_semantics(&local_p, sizeof(local_p));
    set_fork_semantics(&dist_span, sizeof(dist_span));
    set_fork_semantics(&time_steps, sizeof(time_steps));
    set_fork_semantics(&num_threads, sizeof(num_threads));
    set_fork_semantics(&do_checksum, sizeof(do_checksum));
#endif NO_FORK_SEMANTICS

#ifdef RSIM
MEMSYS_ON;
#endif    

#ifdef WWT
    CREATE_ALL(thread_main);
#else
    { 
	int i;
	for (i = 1; i < num_threads; ++i)
	{
	    { if(fork() == 0) {(thread_main)(); TREEBAR(&final_sync, getpid()); exit(0);}};
	}
    }
#endif

    thread_main();

    {TREEBAR(&final_sync, getpid());};

    {exit(0);};
}
