#line 1 "-"
#line 1 "/home/lagavulin/a/gniady/rsim-1.0/apps/utils/lib/rsim_macros"

#line 120

#line 1 "src/em3d.H"
/*
 * Changes copyright (c) 1992-94 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must  the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 *
 * This source file is part of the Wisconsin Wind Tunnel,
 * originally written by Steven K. Reinhardt and David A. Wood,
 * and subsequently maintained and enhanced by Steven K. Reinhardt,
 * Babak Falsafi, and Alvin R. Lebeck.
 *
 * The TEMPEST version of EM3D was originally written by Steve Reinhardt
 * and subsequently rewritten by Babak Falsafi.
 *
 */

/*
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved.
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO {;} SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS."
 *
 * Author: 			Steve Lumetta
 * Version:			11
 * Creation Date:	Sat Jan 23 12:22:55 1993
 * Filename:		em3d.h
 * History:
 *	SL	1	Sat Jan 23 12:42:25 1993
 *		Prepared for release as example (without proper balancing).
 */

#ifndef _EM3D_H
#define _EM3D_H

#undef EXTERN
#undef INIT
#ifdef _EM3D_MAIN
#define EXTERN
#define INIT(a) = (a)
#else
#define EXTERN extern
#define INIT(a)
#endif

#ifdef BLZ
unsigned long long int get_time(void);
#endif

#if (defined(BLZACC_hwstache) && defined(MAKEGRAPH))
#undef MAKEGRAPH
#undef SINGLE
#endif

#ifdef MAKEGRAPH

#include "cyklos/lock.h"

#define lock_decl(l)    ThreadLock l
#define lock_init(l)    { (l) = UNLOCKED; }

#else
#ifdef BLZ

#include "lock.h"

#define lock_decl(l)    MCS_LOCKDEC(l);
#define lock_init(l)    { MCS_LOCKINIT(l); }
#define acq_lock(l)     mcs_lock(l);
#define rel_lock(l)     mcs_unlock(l);

#else

#define lock_decl(l)\
    volatile int (l);
#define lock_init(l)\
    {FREELOCK(&l);};
#define acq_lock(l) \
    {GETLOCK(&l);};
#define rel_lock(l) \
    {FREELOCK(&l);};

#endif
#endif

/* Data type for nodes */
typedef double DATA;

/* List of nodes dependent on a node */
/* List of nodes dependent on a node */
typedef struct to_node_t
{
    struct node_t *node;
    int thread;
} to_node_t;

typedef struct to_node_list_t 
{
    int length;
    struct to_node_t *pairs;
} to_node_list_t;

typedef struct update_list_t
{
    int length;
    int *threads;
} update_list_t;

/* List of nodes on which a node depends (and coefficients) */
typedef struct from_node_list_t 
{
    int length;
    DATA* coeffs;
    DATA **values;
} from_node_list_t;

#ifdef VALOP
#ifdef SINGLE

ERROR -- Value optimization not supported with user protocols

#endif SINGLE
#endif VALOP

#ifdef RSIM
#define STACHE_BLK_SIZE 32
#endif

/* Node structure */
typedef struct node_t 
{
#ifdef VALOP
    DATA *valuep;
#define node_val(np)   (*((np)->valuep))
#else
    DATA value;
#define node_val(np)	((np)->value)
#endif VALOP

    lock_decl(lock)
    struct update_list_t update_nodes;
    struct to_node_list_t to_nodes;
    struct from_node_list_t from_nodes;

#ifdef VALOP
    char pad[STACHE_BLK_SIZE - 28];
#else
    char pad[STACHE_BLK_SIZE - 32];
#endif VALOP
} node_t;

/* Prototypes of functions used to obtain original graph */
void all_initialize_graph(int my_thread);

/* Prototypes of functions used to calculate solution */
void all_compute(node_t *node_vec, int n_nodes);

#include <stdio.h>

/* Private variables */
EXTERN node_t *h_nodes = NULL;
EXTERN node_t *e_nodes = NULL;
EXTERN DATA   *shadow_values = NULL;
EXTERN char   *val_alloc_start = NULL;
EXTERN int     val_alloc_bytes = 0;
EXTERN int     next_random = 0;

/* Shared variables */
typedef struct 
{
    node_t **h_table;
    node_t **e_table;
    double *check_sum;
    TreeBar check_bar;
    TreeBar print_bar;
    TreeBar main_bar;

} shmem_t;

EXTERN shmem_t *shmem;

/* Parameters */
EXTERN int n_nodes INIT(100);
EXTERN int d_nodes INIT(5);
EXTERN int local_p INIT(95);
EXTERN int dist_span INIT(1);
EXTERN int time_steps INIT(100);
EXTERN int num_threads INIT(32);
EXTERN int do_checksum INIT(0);

#ifndef SINGLE

#define end_first_e_iter(t) \
     {TREEBAR(&(shmem->main_bar), MyNum);}
#define end_first_h_iter(t) \
    {TREEBAR(&(shmem->main_bar), MyNum);}
#define end_iter(t, n, m)   \
    {TREEBAR(&(shmem->main_bar), MyNum);}
#define end_all_iters()      

#endif SINGLE

#undef EXTERN
#undef INIT

/*
 * ceil(n/d)
 */
#ifndef DIV_CEIL
#define DIV_CEIL(n,d)   ( ((n)+(d)-1)/(d) )
#endif

#define NUMBER_OF_PROCESSORS num_threads
#define MyNum                my_thread

#ifdef RSIM
#define TRACE4(a1, a2, a3, a4, a5, a6)
#endif

#endif
