/*
 * Changes copyright (c) 1992-94 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must include the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 *
 * This source file is part of the Wisconsin Wind Tunnel,
 * originally written by Steven K. Reinhardt and David A. Wood,
 * and subsequently maintained and enhanced by Steven K. Reinhardt,
 * Babak Falsafi, and Alvin R. Lebeck.
 *
 * The TEMPEST version of EM3D was originally written by Steve Reinhardt
 * and subsequently rewritten by Babak Falsafi.
 *
 */

/*
 *
 * "Copyright (c) 1993 by Steve Lumetta and The Regents of the University 
 * of California.  All rights reserved."
 *
 * Permission to use, copy, modify, and distribute this software and its
 * documentation for any purpose, without fee, and without written agreement is
 * hereby granted, provided that the above copyright notice and the following
 * two paragraphs appear in all copies of this software.
 * 
 * IN NO EVENT SHALL THE UNIVERSITY OF CALIFORNIA BE LIABLE TO ANY PARTY FOR
 * DIRECT, INDIRECT, SPECIAL, INCIDENTAL, OR CONSEQUENTIAL DAMAGES ARISING OUT
 * OF THE USE OF THIS SOFTWARE AND ITS DOCUMENTATION, EVEN IF THE UNIVERSITY OF
 * CALIFORNIA HAS BEEN ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 * 
 * THE UNIVERSITY OF CALIFORNIA SPECIFICALLY DISCLAIMS ANY WARRANTIES,
 * INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY
 * AND FITNESS FOR A PARTICULAR PURPOSE.  THE SOFTWARE PROVIDED HEREUNDER IS
 * ON AN "AS IS" BASIS, AND THE UNIVERSITY OF CALIFORNIA HAS NO OBLIGATION TO
 * PROVIDE MAINTENANCE, SUPPORT, UPDATES, ENHANCEMENTS, OR MODIFICATIONS.
 *
 * Author: 			Steve Lumetta
 * Version:			4
 * Creation Date:	Sat Jan 23 12:51:07 1993
 * Filename:		make_graph.sc
 * History:
 *	SL	1	Sat Jan 23 12:51:45 1993
 *		Prepared for release as example.
 */

#ifdef WWT
ENV
#else
EXTERN_ENV
#endif

#include <stdio.h>
#include <stddef.h>
#include <memory.h>
#include <assert.h>

#ifdef BLZ
#include "syscalls.h"
#include "fatal_error.h"
#include "cyklos/interface.h"
#include "cyklos/lock.h"
#include "trace.h"
#include "tppi_lib_vm.h"
#endif

#include "em3d.h"
#include "prot.h"

/* These constants determine the nature of the random graph. */
#define NUM_H_NODES n_nodes /* 100 */
#define H_NODE_DEGREE d_nodes /* 5 */
#define H_PERCENT_LOCAL local_p /* 0 */

#define NUM_E_NODES n_nodes /* 100 */
#define E_NODE_DEGREE d_nodes /* 5 */
#define E_PERCENT_LOCAL local_p /* 100 */

#define DIST_SPAN dist_span /* 6 */

/* check_percent returns 1 chance percent of the time--otherwise it 
   returns 0. */
 
int check_percent(int my_thread, int chance)
{
    return (((random()/128)%100) < chance);
}

/* gen_number generates a random number between 0 and limit - 1. */
 
int gen_number(int my_thread, int limit)
{
    return (random()/128)%limit;
}

/* generates a number between -(limit - 1) and (limit - 1) excluding 0 */
 
int gen_signed_number(int my_thread, int limit)
{
    int num = gen_number(my_thread, 2*limit);
    
    return (num < limit ? (num - limit) : (num - limit + 1));
}

/* 
 * generates a destination thread number
 * dest thread is a thread on a remote 100-PERCENT_LOCAL
 * percent of time
 */

 
int gen_dest_thread(int my_thread, int percent_local)
{
    int dest_thread;
    
    if (check_percent(my_thread, percent_local))
    {
	return my_thread;
    }

    if (num_threads == 1)
    {
	return my_thread;
    }
    else if (DIST_SPAN == 0)
    {
	return my_thread;
    }

    dest_thread = (my_thread + num_threads +
		   gen_signed_number(my_thread, DIST_SPAN)) % num_threads;
    
    assert(dest_thread >= 0 && dest_thread <= num_threads);
    
    return dest_thread;
}

#if 0

int gen_dest_thread(int my_thread, int percent_local)
{
    int dest_node, dest_local, my_node, dest_thread;

    if (check_percent(my_thread, percent_local))
    {
	return my_thread;
    }

    my_node = machNodeId(my_thread);
    dest_node = (my_node + num_nodes + 
		 gen_signed_number(my_thread, DIST_SPAN)) % num_nodes;
    dest_local = gen_number(my_thread, num_threads_per_node);

    dest_thread = machId(dest_local, dest_node);

    assert(dest_thread >= 0 && dest_thread <= num_threads);

    return dest_thread;
}

#endif

 
double dbl_random(int my_thread)
{
    return (double)random()/1e9;
}

 
void val_shmalloc_init(int my_thread, int nbytes)
{
    val_alloc_bytes = nbytes;

#ifdef ROUND_ROBIN
    val_alloc_start = G_MALLOC(val_alloc_bytes);
#else

#ifdef LOCAL_ALLOC
    val_alloc_start = (void *)local_shm_sbrk(val_alloc_bytes);
#else

#ifdef USER_PROTOCOL
    val_alloc_start = alloc_val_mem(val_alloc_bytes);

#else

ERROR - Should specify one of ROUND_ROBIN, LOCAL_ALLOC or USER_PROTCOL

#endif USER_PROTOCOL
#endif LOCAL_ALLOC
#endif ROUND_ROBIN

     if (my_thread == 0)
     {
	 printf("Allocating %d bytes\n", nbytes * num_threads);
     }
     assert(val_alloc_start);
     memset(val_alloc_start, 0, val_alloc_bytes);
}

void *val_shmalloc(int my_thread, int nbytes)
{
    void *p = val_alloc_start;

    val_alloc_start += nbytes;
    val_alloc_bytes -= nbytes;

    if (val_alloc_bytes < 0) 
    {
	fprintf(stderr, "val_shmalloc(): out of memory\n");
	fflush(stderr);
	exit(-1);
    }

    return p;
}

/*
 * Same as val_shmalloc, but guarantee specified alignment (must be power of 2)
 */
 
void *val_shmalloc_align(int my_thread, int nbytes, int align)
{
    char *p = val_alloc_start;

    if ((unsigned)p & (align-1))
    {
	int adj = align - ((unsigned)p & (align-1));
	p += adj;
	nbytes += adj;
    }

    val_alloc_start += nbytes;
    val_alloc_bytes -= nbytes;
    if (val_alloc_bytes < 0) 
    {
	fprintf(stderr, "val_shmalloc(): out of memory\n");
	fflush(stderr);
	exit(-1);
    }
    return p;
}


#define val_shmalloc_d(m, n)	val_shmalloc_align(m, n, sizeof(double))


void init_node(int my_thread, node_t *np)
{
    lock_init(np->lock);
    np->from_nodes.length = 0;
#ifdef VALOP
    np->valuep = (DATA *)val_shmalloc(my_thread, sizeof(DATA));
#endif VALOP
    node_val(np) = dbl_random(my_thread);
}

/* all_initialize_graph creates a random graph to be solved. */
void all_initialize_graph(int my_thread)
{
    int i, j, k, dest_thread;
    node_t *cur_node;
    node_t *other_node;

    shadow_values = (DATA *)malloc(NUM_E_NODES * sizeof(DATA));

    val_shmalloc_init(my_thread,
		      (NUM_H_NODES * (sizeof(node_t *) + sizeof(node_t) + 
                                       H_NODE_DEGREE*sizeof (node_t *)))
                      + (NUM_E_NODES * (sizeof(node_t *) + sizeof(node_t)
                                        + E_NODE_DEGREE*sizeof (node_t *)))
                      + ((NUM_E_NODES*H_NODE_DEGREE+NUM_H_NODES*E_NODE_DEGREE)
                         * (2*sizeof(node_t *) + sizeof(DATA) + sizeof(int)))
                      + (NUM_E_NODES + NUM_H_NODES) * sizeof(DATA)
                      + 1000 /* slop */);

    BARRIER(shmem->main_bar, my_thread);
    
    /* First the nodes are allocated and initialized. */

    shmem->h_table[my_thread] = (node_t *)
      val_shmalloc_align(my_thread, NUM_H_NODES*sizeof(node_t), 
			 STACHE_BLK_SIZE);

    h_nodes = shmem->h_table[my_thread];

    assert(h_nodes);

    for (i = 0; i < NUM_H_NODES; i++)
    {
	init_node(my_thread, &h_nodes[i]);
    }

    shmem->e_table[my_thread] = (node_t *)
      val_shmalloc_align(my_thread, NUM_E_NODES*sizeof(node_t), 
			 STACHE_BLK_SIZE);

    e_nodes = shmem->e_table[my_thread];

    assert(e_nodes);

    for (i = 0; i < NUM_E_NODES; i++)
    {
	init_node(my_thread, &e_nodes[i]);
    }

    BARRIER(shmem->main_bar, my_thread);
    
    /* 
     * Next the lists of dependent nodes are filled in and the
     * dependents informed so that the length of the from_nodes
     * lists is known later. 
     */

    for (i = 0, cur_node = h_nodes; i < NUM_H_NODES; ++i, ++cur_node) 
    {
	int elem;

	cur_node->to_nodes.length = H_NODE_DEGREE;
	cur_node->to_nodes.pairs = (to_node_t *)
	    val_shmalloc(my_thread, H_NODE_DEGREE*sizeof(to_node_t));

	for (j = 0; j < H_NODE_DEGREE; j++) 
	{
	    do 
	    {
		dest_thread = gen_dest_thread(my_thread, H_PERCENT_LOCAL);
		
		elem = gen_number(my_thread, NUM_E_NODES);

		assert(elem < NUM_E_NODES);

		other_node = &shmem->e_table[dest_thread][elem];
		
		for (k = 0; k < j; k++)
		{
		    if (other_node == cur_node->to_nodes.pairs[k].node)
		    {
			break;
		    }
		}
	    } while (k < j);

	    cur_node->to_nodes.pairs[j].node = other_node;
	    cur_node->to_nodes.pairs[j].thread = dest_thread;

	    TRACE4(TRACE_PROT4,
		   "inc_from: h_node %x(to_nodes)[%d] writing to %d (%x)\n",
		   cur_node,j,dest_thread,other_node);

	    incFrom(my_thread, dest_thread, other_node);
	}
    }    

    for (i = 0, cur_node = e_nodes; i < NUM_E_NODES; ++i, ++cur_node) 
    {
	int elem;

	cur_node->to_nodes.length = E_NODE_DEGREE;
	cur_node->to_nodes.pairs = (to_node_t *)
	    val_shmalloc(my_thread, E_NODE_DEGREE*sizeof(to_node_t));

	for (j = 0; j < E_NODE_DEGREE; j++) 
	{
	    do 
	    {
		dest_thread = gen_dest_thread(my_thread, E_PERCENT_LOCAL);
							 
		elem = gen_number(my_thread, NUM_H_NODES);

		assert(elem < NUM_H_NODES);

		other_node = &shmem->h_table[dest_thread][elem];
		
		for (k = 0; k < j; k++)
		{
		    if (other_node == cur_node->to_nodes.pairs[k].node)
		    {
			break;
		    }
		}
	    } while (k < j);

	    cur_node->to_nodes.pairs[j].node = other_node;
	    cur_node->to_nodes.pairs[j].thread = dest_thread;

	    TRACE4(TRACE_PROT4,
		   "inc_from: e_node %x(to_nodes)[%d] writing to %d (%x)\n",
		   cur_node,j,dest_thread,other_node);

	    incFrom(my_thread, dest_thread, other_node);

	}
    }    

    /* Make sure everyone is done updating to_nodes pointers */
    waitForUpdates(my_thread);
    
    /* Using the length determined in the last section, space is
       allocated for links to nodes on which a node depends and for
       coefficients for the dependencies. */

    for (i = 0, cur_node = h_nodes; i < NUM_H_NODES; ++i, ++cur_node) 
    {
	cur_node->from_nodes.values = (DATA **)
	    val_shmalloc(my_thread, cur_node->from_nodes.length *
			 sizeof(DATA *));

	cur_node->from_nodes.coeffs = (DATA *)
	    val_shmalloc_d(my_thread, cur_node->from_nodes.length *
			   sizeof(DATA));

	cur_node->from_nodes.length = 0;
    }

    for (i = 0, cur_node = e_nodes; i < NUM_E_NODES; ++i, ++cur_node) 
    {
	cur_node->from_nodes.values = (DATA **)
	    val_shmalloc(my_thread, cur_node->from_nodes.length *
			 sizeof(DATA *));

	cur_node->from_nodes.coeffs = (DATA *)
	    val_shmalloc_d(my_thread, cur_node->from_nodes.length *
			   sizeof(DATA));

	cur_node->from_nodes.length = 0;
    }

    BARRIER(shmem->main_bar, my_thread);

    /* The nodes are then added into the from_nodes lists. */

    for (i = 0, cur_node = h_nodes; i < NUM_H_NODES; ++i, ++cur_node)
    {
	for (j = 0; j < cur_node->to_nodes.length; j++)
	{
	    addFrom(my_thread, cur_node, j, dbl_random(my_thread));
	}
    }

    for (i = 0, cur_node = e_nodes; i < NUM_E_NODES; ++i, ++cur_node)
    {
	for (j = 0; j < cur_node->to_nodes.length; j++)
	{
	    addFrom(my_thread, cur_node, j, dbl_random(my_thread));
	}
    }

    waitForUpdates(my_thread);
}
