#ifdef WWT
ENV
#else
EXTERN_ENV
#endif

#include <memory.h>

#ifdef BLZ
#include "cyklos/lock.h"
#include "cyklos/semaphore.h"
#include "tppi_msg.h"
#include "tppi_thread.h"
#include "int_macros.h"
#include "local_lock.h"
#include "wkmsg.h"
#include "proto.h"
#include "tppi_lib_vm.h"
#include "cyklos/interface.h"
#include "util_inlines.h"
#endif

#include "em3d.h"
#include "prot.h"

void initProt(int my_thread)
{
}

#ifndef MAKEGRAPH

/* inc_from is an atomic procedure which increments the length of the
   from_node list of a node. */
void incFrom(int my_thread, int dest_thread, node_t *a_node)
{
    acq_lock(a_node->lock);
    (a_node->from_nodes.length)++;
    rel_lock(a_node->lock);
}

/* add_from is an atomic procedure which adds a node to the from_node list
   of a node. */
void addFrom(int my_thread, node_t *node, int to, double coeff)
{
    node_t *dest_node = node->to_nodes.pairs[to].node;
    int idx;
    
    acq_lock(dest_node->lock);
    idx = dest_node->from_nodes.length++;
    dest_node->from_nodes.values[idx] = &node_val(node);
    dest_node->from_nodes.coeffs[idx] = coeff;
    rel_lock(dest_node->lock);
}

void waitForUpdates(int my_thread)
{
    BARRIER(shmem->main_bar, my_thread);
}

#else

/*
 * Warning: The following code assumes an in-order network
 */

void recvAck()
{
    TPPI_HANDLER_KEY(int, key)
    threadData[0].ackd++;
}

void recvExpt(PT_HANDLER_KEY(int, key))
{
    TPPI_HANDLER_KEY(int, key)
    TPPI_wksend(src, (void *)recvAck, PT_SERIAL_KEY);
}

/* 
 * inc_from is an atomic procedure which increments the length of the
 * from_node list of a node. 
 */

void recvIncFrom(PT_HANDLER_KEY(node_t *, a_node))
{
    TPPI_HANDLER_KEY(node_t *, a_node)
    /*
     * I better not be taking traps
     * while I am doing this!
     */
    lockThread(&a_node->lock);
    (a_node->from_nodes.length)++;
    unlockThread(&a_node->lock);
}

void incFrom(int my_thread, int owner, node_t *a_node)
{
    int ownerNodeId = machNodeId(owner);

    if (machNodeId(my_thread) == ownerNodeId)
    {
	lockThread(&a_node->lock);
	(a_node->from_nodes.length)++;
	unlockThread(&a_node->lock);
    }
    else
    {
	TPPI_wksend(ownerNodeId, (void *)recvIncFrom, (int)a_node);
    }
}

/*
 * add_from is an atomic procedure which adds a node to the from_node list
 * of a node. 
 */

void recvAddFrom(PT_HANDLER_KEY(node_t *, dest_node))
{
    TPPI_HANDLER_KEY(node_t *, dest_node)
    DATA *valuep = (DATA *)TPPI_recv_W();
    int *coeffp;
    int idx;

    /*
     * I better not be taking a trap here!
     */
    lockThread(&dest_node->lock);
    idx = dest_node->from_nodes.length++;
    dest_node->from_nodes.values[idx] = valuep;
    coeffp = (int *)&dest_node->from_nodes.coeffs[idx];
    *coeffp = TPPI_recv_W();
    *(coeffp + 1) = TPPI_recv_W();
    unlockThread(&dest_node->lock);
}

void addFrom(int my_thread, node_t *node, int to, double coeff)
{
    int owner = node->to_nodes.pairs[to].thread;
    int ownerNodeId = machNodeId(owner);

    if (machNodeId(my_thread) == ownerNodeId)
    {
	node_t *dest_node = node->to_nodes.pairs[to].node;
	int idx;

	lockThread(&dest_node->lock);
	idx = dest_node->from_nodes.length++;
	dest_node->from_nodes.values[idx] = &node_val(node);
	dest_node->from_nodes.coeffs[idx] = coeff;
	unlockThread(&dest_node->lock);

    }
    else
    {
	int *coeffp = (int *)&coeff;

	TPPI_wksend_WWW(ownerNodeId, (void *)recvAddFrom, 
			(int)node->to_nodes.pairs[to].node, 
			(int)&node_val(node), *coeffp, *(coeffp + 1));
    }
}

static bool isWaiting(int myLocalId)
{
    return threadData[myLocalId].ackd != num_nodes - 1;
}

void waitForUpdates(int my_thread)
{
    int node;
    int myLocalId = machLocalId(my_thread);
    int myNodeId = machNodeId(my_thread);

    threadBarrier();
    if (myLocalId == 0)
    {
	for (node = 0; node < num_nodes; node++)
	{
	    if (node != myNodeId)
	    {
		TPPI_wksend(node, (void *)recvExpt, PT_ANY_KEY);
	    }
	}
	TPPI_spin(isWaiting, myLocalId);
	threadData[myLocalId].ackd = 0;
    }
    threadGlobalBarrier();
}

#endif MAKEGRAPH

#ifdef SINGLE

#include "stache.h"
#include "home.h"
#include "tppi_msg.h"

void change_tag_handler(PT_HANDLER_KEY(void *, va))
{
    TPPI_HANDLER_KEY(void *, va)
    int tag = TPPI_get_blk_acc(va);
    if (tag == TPPI_Blk_ReadOnly)
    {
	fatal_error("make_writable: va 0x%x ReadOnly tag\n", va);
    }
    else if (tag == TPPI_Blk_Invalid)
    {
	TPPI_change_blk_acc(va, STACHE_BLK_SIZE, TPPI_Blk_Upgrade_RW); 
    }
}

void make_va_writable(void *va)
{
    if (TPPI_get_page_home(va) != TPPI_self_address)
    {
	TPPI_wksend(TPPI_self_address, change_tag_handler, (int)va);
    }
}

void check_va_readonly(void *va)
{
    int tag = TPPI_get_blk_acc(va);
    if (tag == TPPI_Blk_Invalid)
    {
	fatal_error("va 0x%x is not readonly\n", va);
    }
}

void apply_on_all_nodes(void (*f)(void *va), node_t *nodes, int n_nodes)
{
    int idx, val;

    for (idx = 0; idx < n_nodes; idx++)
    {
	for (val = 0; val < nodes[idx].from_nodes.length; val++)
	{
	    void *va = nodes[idx].from_nodes.values[val];
	    (*f)(va);
	}
    }
}

int notInList(int cacherNodeId, int list[], int length)
{
    int i;

    for (i = 0; i < length; i++)
    {
	if (cacherNodeId == machNodeId(list[i]))
	{
	    return FALSE;
	}
    }
    return TRUE;
}

void build_update_list(int my_thread, node_t *nodes, int n_nodes)
{
    int idx, val;

    for (idx = 0; idx < n_nodes; idx++)
    {
	node_t *cur_node = &nodes[idx];
	cur_node->update_nodes.length = 0;
	cur_node->update_nodes.threads = (int *)
            val_shmalloc(my_thread, d_nodes*sizeof(int));

	for (val = 0; val < cur_node->to_nodes.length; val++)
	{
	    int cacherNodeId = 
		machNodeId(cur_node->to_nodes.pairs[val].thread); 
	    if ((machNodeId(my_thread) != cacherNodeId) &&
		notInList(cacherNodeId,
			  cur_node->update_nodes.threads,
			  cur_node->update_nodes.length))
	    {
		cur_node->update_nodes.threads
		    [cur_node->update_nodes.length++] =
		    cur_node->to_nodes.pairs[val].thread;
	    }
	}
    }
}

void recvVal(PT_HANDLER_KEY(long long int *, valuep))
{
    TPPI_HANDLER_KEY(long long int *, valuep)
    *valuep = TPPI_recv_WW();
}

void updateNodes(int my_thread, node_t *nodes, int n_nodes)
{
    int idx, val;
    
    for (idx = 0; idx < n_nodes; idx++)
    {
	for (val = 0; val < nodes[idx].update_nodes.length; val++)
	{
	    int cacher = nodes[idx].update_nodes.threads[val];
	    int *valuep = (int *)&node_val(&nodes[idx]);
	    TPPI_wksend_WW(machNodeId(cacher), (void *)recvVal,
			   (int)valuep, *valuep, *(valuep + 1));
	}
    }
}

void end_first_e_iter(int my_thread)
{
    threadGlobalBarrier();
    build_update_list(my_thread, e_nodes, n_nodes);
    threadGlobalBarrier();
    apply_on_all_nodes(stache_flush, e_nodes, n_nodes);
    apply_on_all_nodes(make_va_writable, e_nodes, n_nodes);
    threadBarrier();
    if (machLocalId(my_thread) == 0)
    {
	msg_barrier();
    }
    threadGlobalBarrier();
}

void end_first_h_iter(int my_thread)
{
    threadGlobalBarrier();
    build_update_list(my_thread, h_nodes, n_nodes);
    threadGlobalBarrier();
    apply_on_all_nodes(stache_flush, h_nodes, n_nodes);
    apply_on_all_nodes(make_va_writable, h_nodes, n_nodes);
    threadBarrier();
    if (machLocalId(my_thread) == 0)
    {
	msg_barrier();
    }
    threadGlobalBarrier();
    enteringCS(threadProcId());
    updateNodes(my_thread, h_nodes, n_nodes);
    waitForUpdates(my_thread);
}

void end_iter(int my_thread, node_t *nodes, int n_nodes)
{
    Uint64 startTime;

    threadGlobalBarrier();
    startTime = get_time();
    updateNodes(my_thread, nodes, n_nodes);
    threadData[my_thread].update_time += get_time() - startTime;
    startTime = get_time();
    waitForUpdates(my_thread);
    threadData[my_thread].receive_time += get_time() - startTime;
}

void end_all_iters()
{
    exitingCS(threadProcId());
}

#endif SINGLE
