#line 1 "-"
#line 1 "/home/lagavulin/a/gniady/rsim-1.0/apps/utils/lib/rsim_macros"

#line 120

#line 1 "src/main.C"
/*************************************************************************/
/*                                                                       */
/*  Copyright (c) 1994 Stanford University                               */
/*                                                                       */
/*  All rights reserved.                                                 */
/*                                                                       */
/*  Permission is given to use, copy, and modify this software for any   */
/*  non-commercial purpose as long as this copyright notice is not       */
/*  removed.  All other uses, including redistribution in whole or in    */
/*  part, are forbidden without prior written permission.                */
/*                                                                       */
/*  This software is provided with absolutely no warranty and no         */
/*  support.                                                             */
/*                                                                       */
/*************************************************************************/

/*************************************************************************/
/*                                                                       */
/*  SPLASH Ocean Code                                                    */
/*                                                                       */
/*  This application studies the role of eddy and boundary currents in   */
/*  influencing large-scale ocean movements.  This implementation uses   */
/*  dynamically allocated four-dimensional arrays for grid data storage. */
/*                                                                       */
/*  Command line options:                                                */
/*                                                                       */
/*     -nN : Simulate NxN ocean.  N must be (power of 2)+2.              */
/*     -pP : P = number of processors.  P must be power of 2.            */
/*     -eE : E = error tolerance for iterative relaxation.               */
/*     -rR : R = distance between grid points in meters.                 */
/*     -tT : T = timestep in seconds.                                    */
/*     -s  : Print timing statistics.                                    */
/*     -o  : Print out relaxation residual values.                       */
/*     -h  : Print out command line options.                             */
/*                                                                       */
/*  Default: OCEAN -n130 -p1 -e1e-7 -r20000.0 -t28800.0                  */
/*                                                                       */
/*  NOTE: This code works under both the FORK and SPROC models.          */
/*                                                                       */
/*************************************************************************/

#ifdef WWT
{;} 
#else
#include <rsim_apps.h>
TreeBar final_sync;
#line 46
#endif

#include <stdlib.h>
#include <malloc.h>
#include "proto.h"
#include "decs.h"

#define DEFAULT_N      258
#define DEFAULT_P        1
#define DEFAULT_E        1e-7
#define DEFAULT_T    28800.0
#define DEFAULT_R    20000.0
#define UP               0
#define DOWN             1
#define LEFT             2
#define RIGHT            3
#define UPLEFT           4
#define UPRIGHT          5
#define DOWNLEFT         6
#define DOWNRIGHT        7
#define PAGE_SIZE     4096

#include <stdio.h>
#include <math.h>
#include <time.h>

struct multi_struct *multi;

struct global_struct *global;

double ****psi;
double ****psim;
double ***psium;
double ***psilm;
double ***psib;
double ***ga;
double ***gb;
double ****work1;
double ***work2;
double ***work3;
double ****work4;
double ****work5;
double ***work6;
double ****work7;
double ****temparray;
double ***tauz;
double ***oldga;
double ***oldgb;
double *f;
double ****q_multi;
double ****rhs_multi;

struct locks_struct *locks;

struct bars_struct *bars;

void subblock();
void slave();
int log_2(int);
void printerr(char *);

int nprocs = DEFAULT_P;
double h1 = 1000.0;
double h3 = 4000.0;
double h = 5000.0;
double lf = -5.12e11;
double res = DEFAULT_R;
double dtau = DEFAULT_T;
double f0 = 8.3e-5;
double beta = 2.0e-11;
double gpr = 0.02;
int im = DEFAULT_N;
int jm;
double tolerance = DEFAULT_E;
double eig2;
double ysca;
int jmm1;
double pi;
double t0 = 0.5e-4 ;
double outday0 = 1.0;
double outday1 = 2.0;
double outday2 = 2.0;
double outday3 = 2.0;
double factjacob;
double factlap;
int numlev;
int *imx;
int *jmx;
double *lev_res;
double *lev_tol;
double maxwork = 10000.0;

#ifdef RSIM
	int phase = 0;
#endif

struct Global_Private *gp;

double *i_int_coeff;
double *j_int_coeff;
int xprocs;
int yprocs;
int *xpts_per_proc;
int *ypts_per_proc;
int minlevel;
int do_stats = 0;
int do_output = 0;

#if defined(TZERO) || defined(BLZ)
void real_main(argc, argv)
#else
void main(argc, argv)
#endif
int argc;
char *argv[];


{
    int i;
    int j;
    int k;
    int x_part;
    int y_part;
    int d_size;
    int itemp;
    int jtemp;
    double procsqrt;
    int temp = 0;
    double min_total;
    double max_total;
    double avg_total;
    double min_multi;
    double max_multi;
    double avg_multi;
    double min_frac;
    double max_frac;
    double avg_frac;
    int ch;
    extern char *optarg;
    unsigned long long computeend;
    unsigned long long start;
#ifdef RSIM
MEMSYS_OFF;
#endif        
    {long time(long *); (start) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    
    while ((ch = getopt(argc, argv, "n:p:e:r:t:soh")) != -1) 
    {
	switch(ch) 
	{
	  case 'n': im = atoi(optarg);
	    if (log_2(im-2) == -1) 
	    {
		    printerr("Grid must be ((power of 2)+2) in each dimension\n");
		    exit(-1);
	    }
	    break;
	  case 'p': nprocs = atoi(optarg); break;
	  case 'e': tolerance = atof(optarg); break;
	  case 'r': res = atof(optarg); break;
	  case 't': dtau = atof(optarg); break;
	  case 's': do_stats = !do_stats; break;
	  case 'o': do_output = !do_output; break;
	  case 'h': printf("Usage: OCEAN <options>\n\n");
	    printf("options:\n");
	    printf("  -nN : Simulate NxN ocean.  N must be (power of 2)+2.\n");
	    printf("  -pP : P = number of processors.  P must be power of 2.\n");
	    printf("  -eE : E = error tolerance for iterative relaxation.\n");
	    printf("  -rR : R = distance between grid points in meters.\n");
	    printf("  -tT : T = timestep in seconds.\n");
	    printf("  -s  : Print timing statistics.\n");
	    printf("  -o  : Print out relaxation residual values.\n");
	    printf("  -h  : Print out command line options.\n\n");
	    printf("Default: OCEAN -n%1d -p%1d -e%1g -r%1g -t%1g\n",
		   DEFAULT_N,DEFAULT_P,DEFAULT_E,DEFAULT_R,DEFAULT_T);
	    exit(0);
	    break;
	}
    }
    
#ifdef WWT
    INITENV(60000000);
#else
    {{TreeBarInit(&(final_sync), NUMBER_OF_PROCESSORS);};};
#endif
    
    if (log_2(nprocs) == -1) 
    {
	printerr("P must be a power of 2\n");
	exit(-1);
    }

    jm = im;
    printf("\n");
    printf("Ocean simulation with W-cycle multigrid solver\n");
    printf("    Processors                         : %1d\n",nprocs);
    printf("    Grid size                          : %1d x %1d\n",im,jm);
    printf("    Grid resolution (meters)           : %0.2f\n",res);
    printf("    Time between relaxations (seconds) : %0.0f\n",dtau);
    printf("    Error tolerance                    : %0.7g\n",tolerance);
    printf("\n");
    
    xprocs = 0;
    yprocs = 0;
    procsqrt = sqrt((double) nprocs);
    j = (int) procsqrt;
    while ((xprocs == 0) && (j > 0)) 
    {
	k = nprocs / j;
	if (k * j == nprocs) 
	{
	    if (k > j) 
	    {
		xprocs = j;
		yprocs = k;
	    } 
	    else 
	    {
		xprocs = k;
		yprocs = j;
	    }
	}
	j--;
    }
    if (xprocs == 0) 
    {
	printerr("Could not find factors for subblocking\n");
	exit(-1);
    }  
    
    minlevel = 0;
    itemp = 1;
    jtemp = 1;
    numlev = 0;
    minlevel = 0;
    while (itemp < (im-2)) 
    {
	itemp = itemp*2;
	jtemp = jtemp*2;
	if ((itemp/yprocs > 1) && (jtemp/xprocs > 1)) 
	{
	    numlev++;
	}
    }  
    
    if (numlev == 0) 
    {
	printerr("Must have at least 2 grid points per processor in each dimension\n");
	exit(-1);
    }
    
    imx = (int *) shmalloc(numlev*sizeof(int));;
    jmx = (int *) shmalloc(numlev*sizeof(int));;
    lev_res = (double *) shmalloc(numlev*sizeof(double));;
    lev_tol = (double *) shmalloc(numlev*sizeof(double));;
    i_int_coeff = (double *) shmalloc(numlev*sizeof(double));;
    j_int_coeff = (double *) shmalloc(numlev*sizeof(double));;
    xpts_per_proc = (int *) shmalloc(numlev*sizeof(int));;
    ypts_per_proc = (int *) shmalloc(numlev*sizeof(int));;
    
    imx[numlev-1] = im;
    jmx[numlev-1] = jm;
    lev_res[numlev-1] = res;
    lev_tol[numlev-1] = tolerance;
    
    for (i=numlev-2;i>=0;i--) 
    {
	imx[i] = ((imx[i+1] - 2) / 2) + 2;
	jmx[i] = ((jmx[i+1] - 2) / 2) + 2;
	lev_res[i] = lev_res[i+1] * 2;
    }
    
    for (i=0;i<numlev;i++) 
    {
	xpts_per_proc[i] = (jmx[i]-2) / xprocs;
	ypts_per_proc[i] = (imx[i]-2) / yprocs;
    }  
    for (i=numlev-1;i>=0;i--) 
    {
	if ((xpts_per_proc[i] < 2) || (ypts_per_proc[i] < 2)) 
	{
	    minlevel = i+1;
	    break;
	}
    }    
    
    for (i=0;i<numlev;i++) 
    {
	temp += imx[i];
    }
    temp = 0;
    j = 0;
    for (k=0;k<numlev;k++) 
    {
	for (i=0;i<imx[k];i++) 
	{
	    j++;
	    temp += jmx[k];
	}
    }
    
    d_size = nprocs*sizeof(double ***);
    psi = (double ****) shmalloc(d_size);;
    psim = (double ****) shmalloc(d_size);;
    work1 = (double ****) shmalloc(d_size);;
    work4 = (double ****) shmalloc(d_size);;
    work5 = (double ****) shmalloc(d_size);;
    work7 = (double ****) shmalloc(d_size);;
    temparray = (double ****) shmalloc(d_size);;
    
    d_size = 2*sizeof(double **);
    for (i=0;i<nprocs;i++) 
    {
	psi[i] = (double ***) shmalloc(d_size);;
	psim[i] = (double ***) shmalloc(d_size);;
	work1[i] = (double ***) shmalloc(d_size);;
	work4[i] = (double ***) shmalloc(d_size);;
	work5[i] = (double ***) shmalloc(d_size);;
	work7[i] = (double ***) shmalloc(d_size);;
	temparray[i] = (double ***) shmalloc(d_size);;
    }
    
    d_size = nprocs*sizeof(double **);
    psium = (double ***) shmalloc(d_size);;
    psilm = (double ***) shmalloc(d_size);;
    psib = (double ***) shmalloc(d_size);;
    ga = (double ***) shmalloc(d_size);;
    gb = (double ***) shmalloc(d_size);;
    work2 = (double ***) shmalloc(d_size);;
    work3 = (double ***) shmalloc(d_size);;
    work6 = (double ***) shmalloc(d_size);;
    tauz = (double ***) shmalloc(d_size);;
    oldga = (double ***) shmalloc(d_size);;
    oldgb = (double ***) shmalloc(d_size);;
    
    gp = (struct Global_Private *) shmalloc((nprocs+1)*sizeof(struct Global_Private));;
    for (i=0;i<nprocs;i++) 
    {
	gp[i].rel_num_x = (int *) shmalloc(numlev*sizeof(int));;
	gp[i].rel_num_y = (int *) shmalloc(numlev*sizeof(int));;
	gp[i].eist = (int *) shmalloc(numlev*sizeof(int));;
	gp[i].ejst = (int *) shmalloc(numlev*sizeof(int));;
	gp[i].oist = (int *) shmalloc(numlev*sizeof(int));;
	gp[i].ojst = (int *) shmalloc(numlev*sizeof(int));;
	gp[i].rlist = (int *) shmalloc(numlev*sizeof(int));;
	gp[i].rljst = (int *) shmalloc(numlev*sizeof(int));;
	gp[i].rlien = (int *) shmalloc(numlev*sizeof(int));;
	gp[i].rljen = (int *) shmalloc(numlev*sizeof(int));;
	gp[i].multi_time = 0;
	gp[i].total_time = 0;
    }
    
    subblock();
    
    x_part = (jm - 2)/xprocs + 2;
    y_part = (im - 2)/yprocs + 2;
    
    d_size = x_part*y_part*sizeof(double) + y_part*sizeof(double *);
    
    global = (struct global_struct *) shmalloc(sizeof(struct global_struct));;  
    for (i=0;i<nprocs;i++) 
    {
	psi[i][0] = (double **) shmalloc(d_size);;
	psi[i][1] = (double **) shmalloc(d_size);;
	psim[i][0] = (double **) shmalloc(d_size);;
	psim[i][1] = (double **) shmalloc(d_size);;
	psium[i] = (double **) shmalloc(d_size);;
	psilm[i] = (double **) shmalloc(d_size);;
	psib[i] = (double **) shmalloc(d_size);;
	ga[i] = (double **) shmalloc(d_size);;
	gb[i] = (double **) shmalloc(d_size);;
	work1[i][0] = (double **) shmalloc(d_size);;
	work1[i][1] = (double **) shmalloc(d_size);;
	work2[i] = (double **) shmalloc(d_size);;
	work3[i] = (double **) shmalloc(d_size);;
	work4[i][0] = (double **) shmalloc(d_size);;
	work4[i][1] = (double **) shmalloc(d_size);;
	work5[i][0] = (double **) shmalloc(d_size);;
	work5[i][1] = (double **) shmalloc(d_size);;
	work6[i] = (double **) shmalloc(d_size);;
	work7[i][0] = (double **) shmalloc(d_size);;
	work7[i][1] = (double **) shmalloc(d_size);;
	temparray[i][0] = (double **) shmalloc(d_size);;
	temparray[i][1] = (double **) shmalloc(d_size);;
	tauz[i] = (double **) shmalloc(d_size);;
	oldga[i] = (double **) shmalloc(d_size);;
	oldgb[i] = (double **) shmalloc(d_size);;
    }
    f = (double *) shmalloc(im*sizeof(double));;
    
    multi = (struct multi_struct *) shmalloc(sizeof(struct multi_struct));;
    
    d_size = numlev*sizeof(double **);
    if (numlev%2 == 1) 
    {         /* To make sure that the actual data 
		 starts double word aligned, add an extra
		 pointer */
	d_size += sizeof(double **);
    }
    for (i=0;i<numlev;i++) 
    {
	d_size += ((imx[i]-2)/yprocs+2)*((jmx[i]-2)/xprocs+2)*sizeof(double)+
	    ((imx[i]-2)/yprocs+2)*sizeof(double *);
    }
    
    d_size *= nprocs;
    
    if (nprocs%2 == 1) 
    {         /* To make sure that the actual data 
		 starts double word aligned, add an extra
		 pointer */
	d_size += sizeof(double ***);
    }
    
    d_size += nprocs*sizeof(double ***);
    q_multi = (double ****) shmalloc(d_size);;
    rhs_multi = (double ****) shmalloc(d_size);;
    
    locks = (struct locks_struct *) shmalloc(sizeof(struct locks_struct));;
    bars = (struct bars_struct *) shmalloc(sizeof(struct bars_struct));;
    
    {FREELOCK(&locks->idlock);};
    {FREELOCK(&locks->psiailock);};
    {FREELOCK(&locks->psibilock);};
    {FREELOCK(&locks->donelock);};
    {FREELOCK(&locks->error_lock);};
    {FREELOCK(&locks->bar_lock);};
    
    {TreeBarInit(&(bars->iteration), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->gsudn), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->p_setup), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->p_redph), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->p_soln), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->p_subph), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_prini), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_psini), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_onetime), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_phase_1), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_phase_2), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_phase_3), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_phase_4), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_phase_5), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_phase_6), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_phase_7), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_phase_8), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_phase_9), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->sl_phase_10), NUMBER_OF_PROCESSORS);};
    {TreeBarInit(&(bars->error_barrier), NUMBER_OF_PROCESSORS);};
    
    link_all();
    
    multi->err_multi = 0.0;
    i_int_coeff[0] = 0.0;
    j_int_coeff[0] = 0.0;
    for (i=0;i<numlev;i++) 
    {
	i_int_coeff[i] = 1.0/(imx[i]-1);
	j_int_coeff[i] = 1.0/(jmx[i]-1);
    }
    
    /* initialize constants and variables
       
       id is a global shared variable that has fetch-and-add operations
       performed on it by processes to obtain their pids.   */
    
    global->id = 0;
    global->psibi = 0.0;
    pi = atan(1.0);
    pi = 4.*pi;
    
    factjacob = -1./(12.*res*res);
    factlap = 1./(res*res);
    eig2 = -h*f0*f0/(h1*h3*gpr);
    
    jmm1 = jm-1 ;
    ysca = ((double) jmm1)*res ;
    
    im = (imx[numlev-1]-2)/yprocs + 2;
    jm = (jmx[numlev-1]-2)/xprocs + 2;
    

#ifdef NO_FORK_SEMANTICS
    set_fork_semantics(&multi, sizeof(multi));
    set_fork_semantics(&global, sizeof(global));
    set_fork_semantics(&psi, sizeof(psi));
    set_fork_semantics(&psim, sizeof(psim));
    set_fork_semantics(&psium, sizeof(psium));
    set_fork_semantics(&psilm, sizeof(psilm));
    set_fork_semantics(&psib, sizeof(psib));
    set_fork_semantics(&ga, sizeof(ga));
    set_fork_semantics(&gb, sizeof(gb));
    set_fork_semantics(&work1, sizeof(work1));
    set_fork_semantics(&work2, sizeof(work2));
    set_fork_semantics(&work3, sizeof(work3));
    set_fork_semantics(&work4, sizeof(work4));
    set_fork_semantics(&work5, sizeof(work5));
    set_fork_semantics(&work6, sizeof(work6));
    set_fork_semantics(&work7, sizeof(work7));
    set_fork_semantics(&temparray, sizeof(temparray));
    set_fork_semantics(&tauz, sizeof(tauz));
    set_fork_semantics(&oldga, sizeof(oldga));
    set_fork_semantics(&oldgb, sizeof(oldgb));
    set_fork_semantics(&f, sizeof(f));
    set_fork_semantics(&q_multi, sizeof(q_multi));
    set_fork_semantics(&rhs_multi, sizeof(rhs_multi));

    set_fork_semantics(&locks, sizeof(locks));
    set_fork_semantics(&bars, sizeof(bars));

    set_fork_semantics(&nprocs, sizeof(nprocs));
    set_fork_semantics(&h1, sizeof(h1));
    set_fork_semantics(&h3, sizeof(h3));
    set_fork_semantics(&h, sizeof(h));
    set_fork_semantics(&lf, sizeof(lf));
    set_fork_semantics(&res, sizeof(res));
    set_fork_semantics(&dtau, sizeof(dtau));
    set_fork_semantics(&f0, sizeof(f0));
    set_fork_semantics(&beta, sizeof(beta));
    set_fork_semantics(&gpr, sizeof(gpr));
    set_fork_semantics(&im, sizeof(im));
    set_fork_semantics(&jm, sizeof(jm));
    set_fork_semantics(&tolerance, sizeof(tolerance));
    set_fork_semantics(&eig2, sizeof(eig2));
    set_fork_semantics(&ysca, sizeof(ysca));
    set_fork_semantics(&jmm1, sizeof(jmm1));
    set_fork_semantics(&pi, sizeof(pi));
    set_fork_semantics(&t0, sizeof(t0));
    set_fork_semantics(&outday0, sizeof(outday0));
    set_fork_semantics(&outday1, sizeof(outday1));
    set_fork_semantics(&outday2, sizeof(outday2));
    set_fork_semantics(&outday3, sizeof(outday3));
    set_fork_semantics(&factjacob, sizeof(factjacob));
    set_fork_semantics(&factlap, sizeof(factlap));
    set_fork_semantics(&numlev, sizeof(numlev));
    set_fork_semantics(&imx, sizeof(imx));
    set_fork_semantics(&jmx, sizeof(jmx));
    set_fork_semantics(&lev_res, sizeof(lev_res));
    set_fork_semantics(&lev_tol, sizeof(lev_tol));
    set_fork_semantics(&maxwork, sizeof(maxwork));

    set_fork_semantics(&gp, sizeof(gp));

    set_fork_semantics(&i_int_coeff, sizeof(i_int_coeff));
    set_fork_semantics(&j_int_coeff, sizeof(j_int_coeff));
    set_fork_semantics(&xprocs, sizeof(xprocs));
    set_fork_semantics(&yprocs, sizeof(yprocs));
    set_fork_semantics(&xpts_per_proc, sizeof(xpts_per_proc));
    set_fork_semantics(&ypts_per_proc, sizeof(ypts_per_proc));
    set_fork_semantics(&minlevel, sizeof(minlevel));
    set_fork_semantics(&do_output, sizeof(do_output));
    set_fork_semantics(&do_stats, sizeof(do_output));

#endif
#ifdef RSIM
MEMSYS_ON;
#endif    
#ifdef WWT
    CREATE_ALL(slave);
#else
    for (i=1;i<nprocs;i++) 
    {
	{ if(fork() == 0) {(slave)(); TREEBAR(&final_sync, getpid()); exit(0);}};
    }
#endif
    
    if (do_output) 
    {
	printf("                       MULTIGRID OUTPUTS\n");
    }
    
    slave();
    {TREEBAR(&final_sync, getpid());};
    {long time(long *); (computeend) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    
    printf("\n");
    printf("                       PROCESS STATISTICS\n");
    printf("                  Total          Multigrid         Multigrid\n");
    printf(" Proc             Time             Time            Fraction\n");
    printf("    0   %15.0f    %15.0f        %10.3f\n",
	   gp[0].total_time,gp[0].multi_time,
	   gp[0].multi_time/gp[0].total_time);
    
    if (do_stats) 
    {
	min_total = max_total = avg_total = gp[0].total_time;
	min_multi = max_multi = avg_multi = gp[0].multi_time;
	min_frac = max_frac = avg_frac = gp[0].multi_time/gp[0].total_time;
	for (i=1;i<nprocs;i++) 
	{
	    if (gp[i].total_time > max_total) 
	    {
		max_total = gp[i].total_time;
	    }
	    if (gp[i].total_time < min_total) 
	    {
		min_total = gp[i].total_time;
	    }
	    if (gp[i].multi_time > max_multi) 
	    {
		max_multi = gp[i].multi_time;
	    }
	    if (gp[i].multi_time < min_multi) 
	    {
		min_multi = gp[i].multi_time;
	    }
	    if (gp[i].multi_time/gp[i].total_time > max_frac) 
	    {
		max_frac = gp[i].multi_time/gp[i].total_time;
	    }
	    if (gp[i].multi_time/gp[i].total_time < min_frac) 
	    {
		min_frac = gp[i].multi_time/gp[i].total_time;
	    }
	    avg_total += gp[i].total_time;
	    avg_multi += gp[i].multi_time;
	    avg_frac += gp[i].multi_time/gp[i].total_time;
	}
	avg_total = avg_total / nprocs;
	avg_multi = avg_multi / nprocs;
	avg_frac = avg_frac / nprocs;
	for (i=1;i<nprocs;i++) 
	{
	    printf("  %3d   %15.0f    %15.0f        %10.3f\n",
		   i,gp[i].total_time,gp[i].multi_time,
		   gp[i].multi_time/gp[i].total_time);
	}
	printf("  Avg   %15.0f    %15.0f        %10.3f\n",
	       avg_total,avg_multi,avg_frac);
	printf("  Min   %15.0f    %15.0f        %10.3f\n",
	       min_total,min_multi,min_frac);
	printf("  Max   %15.0f    %15.0f        %10.3f\n",
	       max_total,max_multi,max_frac);
    }
    printf("\n");
    
    global->starttime = start;
    printf("                       TIMING INFORMATION\n");
    printf("Start time                        : %16.0f\n",
           (double)global->starttime);
    printf("Initialization finish time        : %16.0f\n",
           (double)global->trackstart);
    printf("Overall finish time               : %16.0f\n",
           (double)computeend);
    printf("Total time with initialization    : %16.0f\n",
           (double)(computeend-global->starttime));
    printf("Total time without initialization : %16.0f\n",
           (double)(computeend-global->trackstart));
    printf("    (excludes first timestep)\n");
    printf("\n");
    
    printf("Initialization = %1.0f\n", 
	   (double)(global->trackstart-global->starttime));
    printf("Computation = %1.0f\n", 
	   (double)(computeend-global->trackstart));

    {exit(0);};
}

int log_2(number)

int number;


{
    int cumulative = 1;
    int out = 0;
    int done = 0;
    
    while ((cumulative < number) && (!done) && (out < 50)) 
    {
	if (cumulative == number) 
	{
	    done = 1;
	} 
	else 
	{
	    cumulative = cumulative * 2;
	    out ++;
	}
    }
    
    if (cumulative == number) 
    {
	return(out);
    } 
    else 
    {
	return(-1);
    }
}

void printerr(s)

char *s;


{
    fprintf(stderr,"ERROR: %s\n",s);
}

