#line 1 "-"
#line 1 "/home/lagavulin/a/gniady/rsim-1.0/apps/utils/lib/rsim_macros"

#line 120

#line 1 "src/tomcatv.C"
/* Changes cipyright (c) 1992-93 by Mark Hill, James Larus, and David Wood for
 * the Wisconsin Wind Tunnel Project.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, and David Wood.
 * Any use of this software must  the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 */

/*
 *  Parallel version of tomcatv.
 *  Author: Alvin R. Lebeck
 *  Date: 5/27/92
 *  Description:
 *      Meshgenerations with thompsons solver
 *      modified serial version of W.Gentzsch Fh Regensburg FRG
 *
 *  Disclaimer:
 *      The readability of this code is not that great, since it
 *      is a translated from the original FORTRAN code.
 *      $Revision: 1.1 $ of $Date: 1998/02/03 22:49:56 $
 */

#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#ifdef WWT
#include "cyklos/interface.h"
#include "syscalls.h"
#include "util_inlines.h"
#include "tppi_thread.h"
#include "tppi_vm.h"
#endif
#ifdef BLZACC_hwstache
#include "hwstache.h"
#endif

/* program defines */

#ifdef WWT
{;} 
#else
#include <rsim_apps.h>
TreeBar final_sync;
#line 52
#endif
#define NUMBER_OF_PROCESSORS Num_Threads
#define MyNum                thread_id 
#define MASTER               0 
#ifndef STATIC
#ifdef SIZE
Error--Can not  SIZE if not specifying STATIC
#endif
#define SIZE Array_Size
#else

#ifdef WWT

/* common values */
/* 1026 */
/* 516 */
/* 258 */

/* Let's try to make this dynamic */

#ifndef SIZE
#define SIZE 1026 /* This is one greater than the Fortran parameter N */
#endif

#else

/* not on the WWT, assuming sequent, sequent can't run at 1026 */
#ifndef SIZE
#define SIZE  130 /* This is one greater than the Fortran parameter N */
#endif

#endif /* WWT */
#endif /* DYNAMIC */


#ifndef BLOCK_SIZE
#define BLOCK_SIZE 32
#endif

#define N (SIZE-1)

#define ALFA 0.1
#define RELFA 0.98
#define EPS 0.5e-7
#define H 1.0/(N-1)
#define REL 2.0/RELFA

#define DIFF  (SIZE % ELSPERBLOCK(TWO_DBLS))
#define PAD  (ELSPERBLOCK(TWO_DBLS) - DIFF)

typedef struct two_dbls
{
    double x,y;
} TWO_DBLS;

/* some macros */
#define xa(i,j) (GS[((i)*(SIZE+PAD))+(j)].x)
#define ya(i,j) (GS[((i)*(SIZE+PAD))+(j)].y)

/* typedefs */

typedef struct max_struct
{
    double rxm,rym;
    int irxm,jrxm,irym,jrym;
    char pad[BLOCK_SIZE];
} Max_Struct;


/* a maximum macro */
#define fmax(A,B) ((fabs(A) > fabs(B)) ? fabs(A) : fabs(B))

/* number of elements per BLOCK */
#define ELSPERBLOCK(element) (BLOCK_SIZE/sizeof(element))

#define ALLOC_SIZE ((sizeof(TWO_DBLS) * SIZE * SIZE) + (2*SIZE*BLOCK_SIZE)) 

/* global variables */
typedef struct shared_vars
{   TreeBar start;	
    volatile double dmax;    /* terminating condition variable */
    double rxm,rym;
    int irxm,jrxm;
    int irym,jrym;

} Shared_Vars;

Shared_Vars *sv;
int *start_row;
Max_Struct *xymax;
TWO_DBLS *GS;

#ifdef RSIM		   
     int phase = 0;
#endif
int temptime;
int Num_Threads = 32;
int Num_Threads_Per_Node = 1;
int Num_Nodes = 32;
int LMAX;
int Array_Size;

/* functions */

#define printMatrix() \
{ \
    int thread, itmp, jtmp; \
    for (thread = 0; thread < Num_Threads; ++thread) \
    { \
        if (thread == thread_id) \
        { \
           for (itmp = start_row[thread_id]; itmp < start_row[thread_id+1]; \
                ++itmp) \
           { \
               for (jtmp = 0; jtmp < N; ++jtmp) \
               { \
                   printf("x(%d,%d) %f y(%d,%d) %f\n", itmp, jtmp, \
                                xa(itmp, jtmp), itmp, jtmp, ya(itmp, jtmp)); \
               } \
           } \
         } \
         {TREEBAR(&(sv->start), MyNum);}; \
     } \
}

void sumXYMatrix(int thread_id)
{
    int thread, itmp, jtmp;
    double sumX = 0.0, sumY = 0.0;
    for (itmp = start_row[thread_id]; itmp < start_row[thread_id+1];
	 ++itmp)
    {
	for (jtmp = 0; jtmp < N; ++jtmp)
	{
	    sumX += xa(itmp, jtmp);
	    sumY += ya(itmp, jtmp);
	}
    }
    for (thread = 0; thread < Num_Threads; ++thread)
    {
        if (thread == thread_id)
        {
	    printf("%d: sumX %f sumY %f\n", thread_id, sumX, sumY);
	}
	{TREEBAR(&(sv->start), MyNum);};
    }
}

void printMax(int thread_id)
{
    int thread;
    for (thread = 0; thread < Num_Threads; ++thread)
    {
        if (thread == thread_id)
        {
	    printf("%d: rxm %f rym %f\n", thread_id,
			 xymax[thread_id].rxm, xymax[thread_id].rym);
	}
	{TREEBAR(&(sv->start), MyNum);};
    }
}

void do_work();

int start_time;

#if defined(BLZ) || defined(TZERO)
void real_main
#else
void main
#endif BLZ
(int argc, char **argv)
{
    int i;
    int pid;
    int AllocSize;
    int RowsPerProc;
    int intxymax;
#ifdef NO_FORK_SEMANTICS
    set_fork_semantics(&GS, sizeof(GS));
    set_fork_semantics(&xymax, sizeof(xymax));
    set_fork_semantics(&start_row, sizeof(start_row));
    set_fork_semantics(&sv, sizeof(sv));
    set_fork_semantics(&LMAX, sizeof(LMAX));
    set_fork_semantics(&Num_Threads, sizeof(Num_Threads));
    set_fork_semantics(&Num_Threads_Per_Node, sizeof(Num_Threads_Per_Node));
    set_fork_semantics(&Array_Size, sizeof(Array_Size));
#endif

#ifdef RSIM
MEMSYS_OFF;
#endif    


    if (argc != 4)
    {
	printf("usage: tomcatv arraysize numiterations #of processors\n");
    }

    Num_Threads_Per_Node = 1;
#ifdef RSIM
    Num_Nodes = atoi(argv[3]);
    Num_Threads_Per_Node = 1;
#else
    Num_Nodes = numNodes();
    Num_Threads_Per_Node = TPPI_num_threads;
#endif
    Num_Threads = Num_Nodes * Num_Threads_Per_Node;
    Array_Size = atoi(argv[1]);
    if (Array_Size % Num_Threads != 0)
    {
	printf("arraysize %d is not divisable by %d\n", 
		    Array_Size, Num_Threads);
    }
    Array_Size += 2;
    LMAX = atoi(argv[2]);

#ifdef WWT
    INITENV(ALLOC_SIZE+(sizeof(Shared_Vars)+BLOCK_SIZE)
            +(sizeof(int)*(Num_Threads+1)+BLOCK_SIZE)
            +(sizeof(Max_Struct)*(Num_Threads+1)+BLOCK_SIZE));
#else
    {{TreeBarInit(&(final_sync), NUMBER_OF_PROCESSORS);};};
#endif /* end wwt */

    printf("Running %d iterations.\n", LMAX);

    /* allocate a massive chunk for all the shared data */
    /* this avoids fragmentation */

    AllocSize = ALLOC_SIZE +(sizeof(Shared_Vars)+BLOCK_SIZE) +
      (sizeof(int)*(Num_Threads+1)+ BLOCK_SIZE) +
	(sizeof(Max_Struct)*(Num_Threads+1)+BLOCK_SIZE);

    if ((AllocSize / Num_Nodes) > 50000000)
    {
	printf("Per-node memory requested may not be > 50000000\n");
    }

    printf("requesting %d bytes.\n", AllocSize);
    {long time(long *); (start_time) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;

    GS = (TWO_DBLS *) shmalloc(AllocSize);;

    printf("allocated %d bytes at 0x%x.\n", AllocSize, (int)GS);

    sv = (Shared_Vars *) (((char *) GS)+ALLOC_SIZE);
    start_row = (int *) (((char *) GS)+
			 ALLOC_SIZE+(sizeof(Shared_Vars)+BLOCK_SIZE));
    xymax = (Max_Struct *) (((char *) GS)+
			    ALLOC_SIZE+(sizeof(Shared_Vars)+BLOCK_SIZE)+
			    (sizeof(int)*(Num_Threads+1)+BLOCK_SIZE));

    {TreeBarInit(&(sv->start), NUMBER_OF_PROCESSORS);};
    intxymax = (int) xymax;
    if (intxymax % BLOCK_SIZE)
    	intxymax = ( (int)(intxymax / BLOCK_SIZE) + 1) * BLOCK_SIZE;
    xymax = (Max_Struct *)intxymax;
/* Set up the Profile Labels */

    /* checkout the starting row info */

    RowsPerProc = N / Num_Threads;

    printf("%d threads(s) on %d cpu(s) %d node(s) ~%d rows each.\n",
           Num_Threads_Per_Node, 1, Num_Nodes, 
           RowsPerProc);

    start_row[0] = 2;
    for (pid = 1; pid < Num_Threads; ++pid)
    {
        start_row[pid] = start_row[pid-1] + RowsPerProc;
    }
    start_row[Num_Threads] = N;


    for (i = 1; i <= N; ++i)
    {
        xa(i,1) = (double) (i - 1) / (double) (N-1);
    }

    for (i = 1; i <= N; ++i)
    {
        xa(i,N) = xa(i,1);
        xa(1,i) = 0.0;
        xa(N,i) = 1.0;
        ya(i,1) = 0.0;
        ya(1,i) = xa(i,1);
        ya(N,i) = ALFA*xa(i,1);
    }

    for (i = 1; i <= N; ++i)
    {
        ya(i,N) = (1.0 - xa(i,1)) * ya(1,N) + xa(i,1) * ya(N,N);
    }

    /* spawn the child processes */

#ifdef RSIM
MEMSYS_ON;
#endif    

#ifdef WWT
    CREATE_ALL(do_work);
#else
    for (i=1;i<NUMBER_OF_PROCESSORS;i++) 
    {
	{ if(fork() == 0) {(do_work)(); TREEBAR(&final_sync, getpid()); exit(0);}};
    }
#endif

    do_work();

    {TREEBAR(&final_sync, getpid());};

} /* end main */



void do_work()
{
    unsigned long long init_time, local_start_time;
    int thread_id;
    int local_size;
    int i,j,k,t,i1p,j1p,i2m,j2m,ll,ixcm,jxcm,iycm,jycm;
    int ip,im,jp,jm,l,m = 0,pid;
    double dxcm,dycm,r;
    double xx,yx,xy,yy,a,b,c,qi,qj,pxx,qxx,pyy,qyy,pxy,qxy;
    double **rx,**ry,**d, **aa, *dd;
    unsigned localHeap;

#ifdef RSIM
    thread_id = getpid();
    {TREEBAR(&(sv->start), MyNum);};
#else
/* make sure everyone starts running before assigning ID's */
    {TREEBAR(&(sv->start), MyNum);};
    thread_id = threadsvId();
#endif

#ifdef BLZACC_hwstache
    if (threadId() == 0)
    {
	mark_pages_for_migration();
    }
    {TREEBAR(&(sv->start), MyNum);};
#endif

    /* allocate local arrays, this is done dynamically so
       we can vary the number of processors and the data size accordingly. */

    /* the + 4 is slop because of the fortran to C port */
    local_size = start_row[thread_id+1] - start_row[thread_id] + 4;
    
    if (thread_id == 0)
    {
	printf("Allocating %d local memory\n", 
	       (int)(sizeof(double) * (4 * local_size) +
		     sizeof(double) * 
		     (1 + (4 * local_size)) * SIZE + 
		     256 /* a little padding */)); 
    }

    localHeap = (unsigned)malloc(sizeof(double) * (4 * local_size) +
				 sizeof(double) * 
				 (1 + (4 * local_size)) * SIZE + 
				 256 /* a little padding */);
    
    localHeap += 64;

    rx = (double **) localHeap;
    localHeap += sizeof(double)*local_size;

    ry = (double **) localHeap;
    localHeap += sizeof(double)*local_size;
    
    d  = (double **) localHeap;
    localHeap += sizeof(double)*local_size;

    aa = (double **) localHeap;
    localHeap += sizeof(double)*local_size;
    
    dd = (double *) localHeap;
    localHeap += sizeof(double)*SIZE;
    
    for (i=0;i<local_size;i++)
    {
	rx[i] = (double *) localHeap;
	localHeap += sizeof(double)*SIZE;

	ry[i] = (double *) localHeap;
	localHeap += sizeof(double)*SIZE;

	d[i] = (double *) localHeap;
	localHeap += sizeof(double)*SIZE;

	aa[i] = (double *) localHeap;
	localHeap += sizeof(double)*SIZE;
    }
    
    if (thread_id == 0)
    {
	printf("*** 2-D ITERATION BEHAVIOR ***\n");
	printf("%2s %1s %1s %8s %1s %1s %8s %4s %4s %8s %4s %4s %8s\n",
	       "ITER","I","J","X-COR","I","J","Y-COR",
	       "I","J","X-RES","I","J","Y-RES");
    }

    i1p = 2;
    j1p = 2;
    i2m = N-1;
    j2m = N-1;
    
    ll = 0;
    
    /* finish initialization */

    for (i = start_row[thread_id]; i < start_row[thread_id+1]; ++i)
    {
        for (j = 2; j <= N-1; ++j)
        {
            xa(i,j) = 0.9 * xa(i,1);
            ya(i,j) = 0.9*((1.0-xa(i,1)) *ya(1,j)+ xa(i,1)* ya(N,j));
        }
    } /* end for i = start_row */
    {long time(long *); (temptime) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
    init_time = temptime - start_time;

#ifdef TZERO
    threadBarrier();
    if (threadId() == 0)
    {
	dumpStats("Init");
    }
    threadBarrier();
#endif TZERO
#ifdef RSIM		   
    endphase(phase);
    if (MyNum == MASTER)
     	StatClearAll();
     newphase(++phase);
#endif
   {long time(long *); (local_start_time) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;

    /* make sure all initialization is done */
    {TREEBAR(&(sv->start), MyNum);};

    while (((ll < LMAX) && (sv->dmax > EPS)) || (ll == 0))
    {
        ixcm = jxcm = iycm = jycm = 0;
        dxcm = dycm = 0.0;

        xymax[thread_id].rxm = 0.0;
        xymax[thread_id].rym = 0.0;
        xymax[thread_id].irxm = 0;
        xymax[thread_id].jrxm = 0;
        xymax[thread_id].irym = 0;
        xymax[thread_id].jrym = 0;


	/****************************************
	  the algorithm is as follows:
	  
	  if we are at our first row, then
	  add corrections to the boundary rows.   
	  else
	  add corrections to the rows as we need use them.
	  *******************************************/
	
        for (t=0, i = start_row[thread_id]; i < start_row[thread_id+1]; 
	     ++i,++t)
        {
            ip = i+1;
            im = i-1;
	    
            /* if not first time through add */
            /* corrections to the new column */
	    
            if (ll > 0)
            {
		
	        if (i == start_row[thread_id])
	        {
		    {TREEBAR(&(sv->start), MyNum);};

		    l = 0;
		    for (j = j1p; j <= j2m; ++j)
		    {
			++l;
			xa(i,j) = xa(i,j) + rx[t][l];
			ya(i,j) = ya(i,j) + ry[t][l];

			xa(start_row[thread_id+1]-1,j) =
                           xa(start_row[thread_id+1]-1,j) +
			   rx[start_row[thread_id+1]-
			      start_row[thread_id]-1][l];

			ya(start_row[thread_id+1]-1,j) =
                           ya(start_row[thread_id+1]-1,j) +
			   ry[start_row[thread_id+1]-
			      start_row[thread_id]-1][l];
		    }

	    /* check in the boundary rows we just to updated */
	    /* turn around and prefetch the same row */
                    
		    {TREEBAR(&(sv->start), MyNum);};

	        } /* end adding corrections to boundary rows */

		/* add corrections to the next row used */
		/* the last row has already been updated, because the */
		/* other processor needed that value right away. */

		if (i < start_row[thread_id+1]-2)
		{
	            l = 0;
	            for (j = j1p; j <= j2m; ++j)
	            {
	                ++l;
	                xa(ip,j) = xa(ip,j) + rx[t+1][l];
		        ya(ip,j) = ya(ip,j) + ry[t+1][l];
	            }
		}
            } /* end if ll > 0 */

           /* j is inner loop in C */
            m = 0;
            for (j = j1p; j <= j2m; ++j)
            {
                m++;
                jp = j+1;
                jm = j-1;
                xx = xa(ip,j)-xa(im,j);
                yx = ya(ip,j)-ya(im,j);
                xy = xa(i,jp)-xa(i,jm);
                yy = ya(i,jp)-ya(i,jm);
                a  = 0.25 * (xy*xy+yy*yy);
                b  = 0.25*(xx*xx+yx*yx);
                c  = 0.125 *(xx*xy+yx*yy);
                qi = 0.0;
                qj = 0.0;

                /* as commented out in spec version
                qi = a*0.5;
                qj = b*0.5;
                */

                aa[t][m] = -b;
                dd[m] = b+b+a*REL;
		/*
                dd[t][m] = b+b+a*REL;
		*/
                pxx = xa(ip,j)-2.0*xa(i,j)+xa(im,j);
                qxx = ya(ip,j)-2.0*ya(i,j)+ya(im,j);
                pyy = xa(i,jp)-2.0*xa(i,j)+xa(i,jm);
                qyy = ya(i,jp)-2.0*ya(i,j)+ya(i,jm);
                /*
                pxy = xa(ip,jp)-xa(ip,jm)-xa(im,jp)+xa(im,jm);
                qxy = ya(ip,jp)-ya(ip,jm)-ya(im,jp)+ya(im,jm);
                */
                pxy = xa(ip,jp);
                pxy -= xa(ip,jm);
                pxy -= xa(im,jp);
                pxy += xa(im,jm);
                qxy = ya(ip,jp);
                qxy -= ya(ip,jm);
                qxy -= ya(im,jp);
                qxy += ya(im,jm);
                
/*
     CALCULATE RESIDUALS (EQUIVALENT TO RIGHT HAND SIDES OF EQUS.)
*/

                rx[t][m] = a*pxx+b*pyy-c*pxy+xx*qi+xy*qj;
                ry[t][m] = a*qxx+b*qyy-c*qxy+yx*qi+yy*qj;
                
            }


            for (j = 1; j <= m; ++j)
            {
                if(fabs(rx[t][j]) > fabs(xymax[thread_id].rxm))
                {
                    xymax[thread_id].rxm  = rx[t][j];
                    xymax[thread_id].irxm = i;
                    xymax[thread_id].jrxm = j;
                }


                if(fabs(ry[t][j]) > fabs(xymax[thread_id].rym)) 
                {
                    xymax[thread_id].rym  = ry[t][j];
                    xymax[thread_id].irym = i;
                    xymax[thread_id].jrym = j;
                }
            }

	    /*
            d[t][1] = 1.0/dd[t][1];
	    */
            d[t][1] = 1.0/dd[1];

            for (j = 2; j <= m; ++j)
            {
                r = aa[t][j]*d[t][j-1];
		/*
                d[t][j] = 1.0/(dd[t][j]-aa[t][j-1]*r);
		*/
                d[t][j] = 1.0/(dd[j]-aa[t][j-1]*r);
                rx[t][j] = rx[t][j] - rx[t][j-1]*r;
                ry[t][j] = ry[t][j] - ry[t][j-1]*r;
            }

            rx[t][m] = rx[t][m]*d[t][m];
            ry[t][m] = ry[t][m]*d[t][m];

        } /* end i loop */

	/* BACKWARD WAVE */

        for (t=0,i = start_row[thread_id]; i < start_row[thread_id+1]; ++i,++t)
        {
            k = 0;
            for (j = 2; j <= m ; ++j)
            {
                k = m-j+1;
                rx[t][k] = (rx[t][k]-aa[t][k]*rx[t][k+1])*d[t][k];
                ry[t][k] = (ry[t][k]-aa[t][k]*ry[t][k+1])*d[t][k];
            }
        } /* end backward wave */


        /* make sure that everyone has computed their local maximums */
	{TREEBAR(&(sv->start), MyNum);};

        ++ll; /* increment the loop counter */

        if (thread_id == 0)
        {
            sv->rxm = 0.0;
            sv->rym = 0.0;
            sv->dmax = 0.0;

            for (pid = 0; pid < Num_Threads; ++pid)
            {
                if (fabs(xymax[pid].rxm) > fabs(sv->rxm))
                {
                    sv->rxm = xymax[pid].rxm;
                    sv->irxm = xymax[pid].irxm;
                    sv->jrxm = xymax[pid].jrxm;
                }
                if (fabs(xymax[pid].rym) > fabs(sv->rym))
                {
                    sv->rym = xymax[pid].rym;
                    sv->irym = xymax[pid].irym;
                    sv->jrym = xymax[pid].jrym;
                }
            }

            printf("%2d %1d %1d %8.3e %1d ",
                ll,ixcm,jxcm,dxcm,iycm);
            printf("%1d %8.3e %4d %4d %8.3e ",
                jycm,dycm, sv->irxm, sv->jrxm, sv->rxm);
            printf("%4d %4d %8.3e\n",
                sv->irym, sv->jrym, sv->rym);

            /* update the terminating condition */
            sv->dmax = fmax(sv->dmax,fmax(sv->rxm,sv->rym));

        } /* end if proc = 0 */

        /* make sure everyone sees the new value of dmax */
	{TREEBAR(&(sv->start), MyNum);};
    }

    if (thread_id == 0)
    {
	printf("Initialization = %u\n", (unsigned int)init_time);
	{long time(long *); (temptime) = time(0); }/* if you want clock cycles, you should call sysclocks instead of time (which gives seconds) */;
	printf("Computation = %u\n", 
	       (unsigned int)(temptime - local_start_time));
    }

#ifdef RSIM
   endphase(phase);
   if (MyNum == MASTER)
	StatReportAll();
#endif

#ifdef TZERO
    threadBarrier();
    if (threadId() == 0)
    {
	dumpStats("End");
    }
    threadBarrier();
#endif TZERO
}
