#ifndef _common_h
#define _common_h
#include "tppi_fatal_error.h"

extern int node_number;

#ifndef MIN
#define MIN(a,b)        ((a) < (b) ? (a) : (b))
#endif

#ifndef MAX
#define MAX(a,b)        ((a) > (b) ? (a) : (b))
#endif

#ifndef NULL
#define NULL	0
#endif

#ifndef TRUE
#define TRUE	1
#define FALSE	0
#endif

/*
 * ceil(n/d)
 */
#ifndef DIV_CEIL
#define DIV_CEIL(n,d)   ( ((n)+(d)-1)/(d) )
#endif

#ifdef NO_ASSERTIONS
#define assert(B)
#else
#ifndef assert
#define assert(B)  							\
{									\
    if (!(B))								\
	fatal_error("assertion failed at line %d of file %s: %s\n",	\
		    __LINE__, __FILE__, #B);			\
}
#endif
#endif


/*
 * xmalloc() calls malloc() and returns only if malloc succeeds.
 * If malloc() fails, fatal_error() is called.
 */
extern void *xmalloc(unsigned int nbytes);

/*
 * xmalloc_z() calls xmalloc(), then memset().
 */
extern void *xmalloc_z(unsigned int nbytes);


/*
 * new_vec(T,n) allocates an array of 'n' items of type 'T' (using xmalloc).
 */
#define new_vec(T, n)	((T *)xmalloc((n) * sizeof(T)))

/*
 * new_vec_z(T,n) uses xmalloc_z, handy w/multinde for allocating things
 * that used to be in bss.
 */
#define new_vec_z(T, n)	((T *)xmalloc_z((n) * sizeof(T)))

#endif
