/* Copyright (c) 1995 by Mark Hill, James Larus, and David Wood for 
 * the Wisconsin Wind Tunnel Project and Joel Saltz for the High Performance 
 * Software Laboratory, University of Maryland, College Park.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, Joel Saltz, and
 * David Wood.  Any use of this software must include the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 * Shubhendu Mukherjee (Shubu) and Shamik Sharma generated the sequential 
 * version in C. 
 * Shubhendu Mukherjee (Shubu) parallelized unstructured on Tempest. 
 */

/*
 * Joint project between the Wisconsin Wind Tunnel at the University of Wisconsin, 
 * Madison and Joel Saltz's research group at the University of Maryland, College 
 * Park.  
 *
 * Participants: Shubu Mukherjee (*) and Shamik Sharma (!).
 *               Mark Hill (*), Jim Larus(*), Anne Rogers (%), Joel Saltz (!).
 *		 
 *		 (*) University of Wisconsin-Madison
 *		 (!) University of Maryland, College Park
 *		 (%) Princeton University
 */

#ifndef  _unstructured_h
#define  _unstructured_h

#ifdef MAIN_UNS
MAIN_ENV
#else
EXTERN_ENV
#endif

#define CLOCK_T unsigned long long
#define CLOCK_HZ 1

#ifdef FIRST_COORD_BISEC
#define COORD_BISEC
#endif

#ifdef RENAME
#define RENAME_NODES
#define RENAME_EDGES
#define RENAME_FACES
#endif

#include <stdio.h>
#include <math.h>
#include <string.h>
#include <assert.h>

#include "combination.h"
#include "update_macros.h"

#ifdef USE_ACK_UPDATES
#include "update.h"
#endif

#define MyNum                myproc
#define NUMBER_OF_PROCESSORS nprocs

#define min(a,b)        (((a) < (b)) ? (a) : (b))
#define max(a,b)        (((a) > (b)) ? (a) : (b))
#define sqr(a)          ((a) * (a))
#define abs(a)		(((a) > 0) ? (a) : (-(a)))

	  
#ifndef DIV_CEIL
#define DIV_CEIL(n,d)   ( ((n)+(d)-1)/(d) )
#endif

#ifdef MESH2K

#define MAX_NODES           10000
#define MAX_EDGES           20000
#define MAX_FACES            3000
#define MAX_BOUNDARY_NODES   2000
#define MAX_BOUNDARY_TYPES     10

#else

#define MAX_NODES          100000
#define MAX_EDGES          360000
#define MAX_FACES           25000
#define MAX_BOUNDARY_NODES  12000
#define MAX_BOUNDARY_TYPES     10

#endif

#define DEFAULT_BND_EDGE_DATA 0.5
#define TYPE_TAN 3
#define TYPE_IO  6

#define one_d_array(arrayname, index)   arrayname[(index) - 1]

#define two_d_array(arrayname, i, j, dim_i) \
        arrayname[((i) - 1) + ((j) - 1) * dim_i]

#define three_d_array(arrayname, i, j, k, dim_i, dim_j) \
        arrayname[((i) - 1) + \
                  ((j) - 1) * dim_i + \
                  ((k) - 1) * dim_i * dim_j]

#define four_d_array(arrayname, i, j, k, l, dim_i, dim_j, dim_k) \
        arrayname[((i)-1) + \
                  ((j)-1) * dim_i + \
                  ((k)-1) * dim_i * dim_j + \
                  ((l)-1) * dim_i * dim_j * dim_k]

EXTERN int   nprocs;
EXTERN int   myproc;
EXTERN int   numBoundaryNodes;
EXTERN int   numBoundaryEdges;

EXTERN int   *BOUNDARYFACEPTR;
#define      boundaryFacePtr(i, j)  	two_d_array(BOUNDARYFACEPTR, j, i, 3)

EXTERN int   *BOUNDARYNODEPTR;
#define      boundaryNodePtr(i, j)  	two_d_array(BOUNDARYNODEPTR, j, i, 3)

EXTERN int   *BOUNDARYEDGEPTR;
#define      boundaryEdgePtr(i, j)  	two_d_array(BOUNDARYEDGEPTR, j, i, 3)

EXTERN int   numMeshNodes;

#ifdef RENAME_NODES
EXTERN int   *TMPISBOUNDARYNODE;
#define      TmpisBoundaryNode(i) 	one_d_array(TMPISBOUNDARYNODE,i)
#endif

EXTERN int   *ISBOUNDARYNODE;
#define      isBoundaryNode(i) 	one_d_array(ISBOUNDARYNODE,i)

EXTERN double *NODE_COORD_X;
EXTERN double *NODE_COORD_Y;
EXTERN double *NODE_COORD_Z;

#define node_coord_x(i)		one_d_array(NODE_COORD_X, i)
#define node_coord_y(i)		one_d_array(NODE_COORD_Y, i)
#define node_coord_z(i)		one_d_array(NODE_COORD_Z, i)

EXTERN int   *BOUNDARYNODE;

#ifdef RENAME_NODES

#define      boundaryNode(i)   	one_d_array(BOUNDARYNODE,i)

#else

#define      boundaryNode(i)   	one_d_array(BOUNDARYNODE,node_indirection_array(i))

#endif RENAME_NODES

EXTERN float  *BOUNDARYNODEDATA;

#ifdef RENAME_NODES

#define      bndNodeData(i)    	one_d_array(BOUNDARYNODEDATA,i)

#else

#define      bndNodeData(i)    	one_d_array(BOUNDARYNODEDATA,node_indirection_array(i))

#endif RENAME_NODES

EXTERN int max_allocated_nodes;


EXTERN int *NODE_INDIRECTION_ARRAY;
#define node_indirection_array(i)	one_d_array(NODE_INDIRECTION_ARRAY,i)

EXTERN int *NDX;
#define ndx(i)				one_d_array(NDX, i)

EXTERN unsigned *START_ADDRESS;
#define start_address(i)		one_d_array(START_ADDRESS, (i+1))

EXTERN float  *FORCES;		/* local with tempest reductions */
EXTERN float  *VEL_DELTA;
EXTERN float  *VELOCITY;
EXTERN float  *TMPA;
EXTERN float  *TMPB;

EXTERN float  *OUTPUT;
/* 
 * Size of single element in forces, vel_delta, velocity, tmpA, and tmpB
 * in bytes. 
*/

#ifdef PADDING16

#define NUM_IN_NODE		4

#else
#ifdef PADDING32

#define NUM_IN_NODE		8

#else

#define NUM_IN_NODE		3

#endif
#endif

#define SIZE_OF_NODE		(sizeof(float) * NUM_IN_NODE)
#define BASE_SIZE_OF_NODE	(sizeof(float) * 3)

#ifdef RENAME_NODES

#define      forces(i, j)      	two_d_array(FORCES,j,i,NUM_IN_NODE)
#define      vel_delta(i, j)  	two_d_array(VEL_DELTA,j,i,NUM_IN_NODE)
#define      velocity(i, j)  	two_d_array(VELOCITY,j,i,NUM_IN_NODE)
#define      tmpA(i, j)     	two_d_array(TMPA,j,i,NUM_IN_NODE)
#define      tmpB(i, j)     	two_d_array(TMPB,j,i,NUM_IN_NODE)

#else

#define      forces(i, j)      	two_d_array(FORCES,j,node_indirection_array(i),NUM_IN_NODE)
#define      vel_delta(i, j)  	two_d_array(VEL_DELTA,j,node_indirection_array(i),NUM_IN_NODE)
#define      velocity(i, j)  	two_d_array(VELOCITY,j,node_indirection_array(i),NUM_IN_NODE)
#define      tmpA(i, j)     	two_d_array(TMPA,j,node_indirection_array(i),NUM_IN_NODE)
#define      tmpB(i, j)     	two_d_array(TMPB,j,node_indirection_array(i),NUM_IN_NODE)

#endif RENAME_NODES

EXTERN float  *LOCAL_FORCES;
EXTERN float  *LOCAL_VEL_DELTA;
EXTERN float  *LOCAL_TMPA;
EXTERN float  *LOCAL_TMPB;

/* 
 * These should have indirection arrays because of the channel reduction, 
 * which requires same alignment of local arrays as the global ones.  Of 
 * course, this can be easily changed.			-Shubu /SSM/
 */

#ifdef RENAME_NODES

#define      local_forces(i, j)  	\
                      two_d_array(LOCAL_FORCES,j, i, NUM_IN_NODE)

#define      local_vel_delta(i, j)  	\
                      two_d_array(LOCAL_VEL_DELTA,j,i, NUM_IN_NODE)

#define      local_tmpA(i, j)     	\
                      two_d_array(LOCAL_TMPA, j, i, NUM_IN_NODE)

#define      local_tmpB(i, j)     	\
                      two_d_array(LOCAL_TMPB, j, i, NUM_IN_NODE)

#else

#define      local_forces(i, j)  	\
                      two_d_array(LOCAL_FORCES,j,node_indirection_array(i), NUM_IN_NODE)

#define      local_vel_delta(i, j)  	\
                      two_d_array(LOCAL_VEL_DELTA,j,node_indirection_array(i), NUM_IN_NODE)

#define      local_tmpA(i, j)     	\
                      two_d_array(LOCAL_TMPA, j, node_indirection_array(i), NUM_IN_NODE)

#define      local_tmpB(i, j)     	\
                      two_d_array(LOCAL_TMPB, j, node_indirection_array(i), NUM_IN_NODE)

#endif RENAME_NODES

EXTERN float  *VEL_BOUNDARY;
#define      vel_boundary(i, j)  two_d_array(VEL_BOUNDARY,j,i,3)

EXTERN float  *VEL_BACKUP;
#define      vel_backup(i, j)  	two_d_array(VEL_BACKUP, j, i, 3)

EXTERN int   numMeshEdges;

struct edge_t
{
    int node1, node2, is_bnd;
    float data1, data2, data3;
};

EXTERN struct edge_t *EDGE;    /* read-only */
#define      edge(i)        	one_d_array(EDGE, i)

EXTERN int   *TMPISBOUNDARYEDGE;	 	     /* read-only */
#define      TmpisBoundaryEdge(i) 	one_d_array(TMPISBOUNDARYEDGE,i)

EXTERN int   *TMPEDGE;                      /* read-only */
#define      tmp_edge(i, j)    	two_d_array(TMPEDGE, j, i, 2)

EXTERN int   *BOUNDARYEDGE;
#define      boundaryEdge(i)   one_d_array(BOUNDARYEDGE,i)

EXTERN float  *BOUNDARYEDGEDATA;
#define      bndEdgeData(i)    one_d_array(BOUNDARYEDGEDATA,i)

EXTERN int  numMeshFaces;

EXTERN int  *TMPFACE;             /* read-only */
#define     tmp_face(i,j)          two_d_array(TMPFACE, j, i, 4)

struct face_t
{
    int node1, node2, node3, node4, is_bnd;
    float data1, data2, data3;
};

EXTERN struct face_t  *FACE;             /* read-only */
#define     face(i)          one_d_array(FACE, i)

EXTERN int  *BOUNDARYFACE;
#define     boundaryFace(i)    one_d_array(BOUNDARYFACE,i)

EXTERN int  *TMPISBOUNDARYFACE;	/* read-only, provides type */
#define     TmpisBoundaryFace(i)  one_d_array(TMPISBOUNDARYFACE, i)

EXTERN int low_node, high_node, low_edge, high_edge, low_face, high_face;
EXTERN int low_bnd_node, high_bnd_node;

EXTERN int numTimeSteps;
EXTERN int *face_to_proc;

#ifdef RENAME_NODES
EXTERN int *TMP_NODE_TO_PROC;
#define tmp_node_to_proc(i)		one_d_array(TMP_NODE_TO_PROC,i)
#endif

EXTERN int *NODE_TO_PROC;
#define node_to_proc(i)		one_d_array(NODE_TO_PROC,i)

#define MAX_PROC	32

#define MAX_NODES_PER_PROC	(MAX_NODES / 16)

EXTERN struct node_list_t
{
    int num;
    int nodes[MAX_NODES_PER_PROC];
} *node_list;

#define MAX_BND_NODES_PER_PROC	(MAX_BOUNDARY_NODES / 16)

EXTERN struct bnd_node_list_t
{
    int num;
    int nodes[MAX_BND_NODES_PER_PROC];
} *bnd_node_list;

EXTERN int current_node[MAX_PROC];
EXTERN int current_bnd_node[MAX_PROC];

#define MAX_EDGES_PER_PROC	(MAX_EDGES / 16)

EXTERN struct edge_list_t
{	
    int num;
    int edges[MAX_EDGES_PER_PROC];
}  *edge_list;

EXTERN int current_edge[MAX_PROC];

struct block_t
{
    int low, high;
};

#ifdef RENAME_EDGES
EXTERN struct block_t *edge_block;
EXTERN int *EDGE_MAP;
#define edge_map(i)	one_d_array(EDGE_MAP,i)
#endif

#ifdef RENAME_NODES
EXTERN struct block_t *node_block;
#endif

#ifdef RENAME_FACES
EXTERN struct block_t *face_block;
#define face_map(i)	one_d_array(FACE_MAP,i)
EXTERN int *FACE_MAP;
void rename_faces();
#endif

EXTERN int MAX_FACES_PER_PROC;

EXTERN struct face_list_t
{	
    int num;
    int *faces;
}  *face_list, *bnd_face_list;

EXTERN int current_face[MAX_PROC];
EXTERN int current_bnd_face[MAX_PROC];

struct global_t
{
    float vel[MAX_PROC];
    float v_tot[MAX_PROC];
#ifndef NO_LOCKING
    ALOCKDEC(lock, (MAX_NODES+1))
#endif
    BARDEC(barvar)
} *global;

#define force_lock	(global->lock)
#define vel_lock	(global->lock)
#define tmpA_lock	(global->lock)
#define tmpB_lock	(global->lock)

#ifndef NO_LOCKING
EXTERN char *TOUCHED_REMOTE;
#define     touched_remote(i)  one_d_array(TOUCHED_REMOTE, i)

#ifdef COORD_BISEC
#define IF_TOUCHED_REMOTE(i)	if (touched_remote(i))
#else
#define IF_TOUCHED_REMOTE(i)
#endif

#endif

#define START 1
#define END   2
#define SIZE  3

#define MAX_STACHE_BLK_SIZE    	4096	/* one page */

/* reduction using direct channels */
EXTERN int *send_edge_channel_id, *send_edge_face_channel_id;
EXTERN int *recv_edge_channel_id, *recv_edge_face_channel_id;
EXTERN int *recv_edge_channel_size, *recv_edge_face_channel_size;

/* 
 * This needs to be 2^(x) - 1, probably because of alignment restrictions
 * for send region in fine updates. 
 */
#define MAX_SEND_PER_PROC       1023

EXTERN struct send_list_t
{
    int num;
    int *nodes;
} *send_edge_list, *send_edge_face_list;

EXTERN int recv_reduction_edge_list_num[MAX_PROC];
EXTERN int **recv_reduction_edge_list;

EXTERN int recv_reduction_edge_face_list_num[MAX_PROC];
EXTERN int **recv_reduction_edge_face_list;

/* for fine update */
EXTERN int *send_update_edge_channel_id, *send_update_edge_face_channel_id;
EXTERN int *recv_update_edge_channel_id, *recv_update_edge_face_channel_id;
EXTERN int *recv_update_edge_channel_size, *recv_update_edge_face_channel_size;

EXTERN struct off_proc_node_list_t
{
  int nodes[MAX_SEND_PER_PROC];
  int num;
} *SEND_UPDATE_EDGE_LIST, *SEND_UPDATE_EDGE_FACE_LIST,
  *INIT_UPDATE_EDGE_LIST, *INIT_UPDATE_EDGE_FACE_LIST;


#define init_update_edge_list(proc_send) \
        one_d_array(INIT_UPDATE_EDGE_LIST, (proc_send+1))

#define init_update_edge_face_list(proc_send) \
        one_d_array(INIT_UPDATE_EDGE_FACE_LIST, (proc_send+1))

#define send_update_edge_list(proc_send) \
        one_d_array(SEND_UPDATE_EDGE_LIST, (proc_send+1))

#define send_update_edge_face_list(proc_send) \
        one_d_array(SEND_UPDATE_EDGE_FACE_LIST, (proc_send+1))

typedef struct
{
    float x, y, z;
} recv_element;

EXTERN recv_element send_buffer[MAX_SEND_PER_PROC];

EXTERN recv_element **recv_edge_list, **recv_edge_face_list;
EXTERN recv_element **recv_update_edge_list, **recv_update_edge_face_list;

#define TYPE_EDGE		0
#define TYPE_EDGE_FACE		1  /* means both edge and face */

EXTERN struct holder_t
{	
    float *tmpA;
    float *tmpB;
    float *velocity;
    float *vel_delta;
} *global_address_holder;

#ifdef RENAME_NODES
#define NODE_LOOP(i,ii)	for (i=low_node; i<=high_node;i++)
#else
#define NODE_LOOP(i,ii) \
                        for (ii=0,i=node_list[myproc].nodes[ii]; \
			     ii<node_list[myproc].num; \
			     ii++, i=node_list[myproc].nodes[ii])
#endif

#ifdef RENAME_EDGES
#define EDGE_LOOP(i,ii)	for (i=low_edge; i<=high_edge;i++)
#else
#define EDGE_LOOP(i,ii) \
                        for (ii=0,i=edge_list[myproc].edges[ii]; \
			     ii<edge_list[myproc].num; \
			     ii++, i=edge_list[myproc].edges[ii])
#endif

#ifdef RENAME_FACES
#define FACE_LOOP(i,ii)	for (i=low_face; i<=high_face;i++)
#else
#define FACE_LOOP(i,ii) \
                        for (ii=0,i=face_list[myproc].faces[ii]; \
			     ii<face_list[myproc].num; \
			     ii++, i=face_list[myproc].faces[ii])
#endif

#define BND_NODE_LOOP(b,ii) \
                       for (ii=0,b=bnd_node_list[myproc].nodes[ii]; \
			    ii<bnd_node_list[myproc].num; \
			    ii++, b=bnd_node_list[myproc].nodes[ii])

#define BND_FACE_LOOP(b,ii) \
                       for (ii=0,b=bnd_face_list[myproc].faces[ii]; \
			    ii<bnd_face_list[myproc].num; \
			    ii++, b=bnd_face_list[myproc].faces[ii])

#ifdef VERBOSE
#define verbose(s)		\
{				\
    if (myproc == 0)	\
     {				\
	printf("%s", s);	\
        fflush(stdout);		\
     }				\
}
#else
#define verbose(s)
#endif

#ifdef SHM_REDUCTION
#define reduction(a,b,c) shm_reduction(a,b)
#else
#ifdef BULK_REDUCTION
#define reduction(a,b,c) bulk_reduction(a,b)
#else 
#ifdef DIRECT_REDUCTION
#define reduction(a,b,c) direct_reduction(a,b,c)
#else
#define reduction(a,b,c) 
#endif DIRECT_REDUCTION
#endif BULK_REDUCTION
#endif SHM_REDUCTION

void ComputeForces(float, int);
void ComputeVel();
void ComputeVelocityChange();
void SmoothenVelocity(int);
void Output();
void Sweep(int);
void Initialize();
void SetBoundaryData();
void SortBoundaryNodesFacesBasedOnType();
void ReadMesh(char *, int *, int *, int *, int *, int *);
void PrintParameters();
void print_initial_message(char *, int);
int  low_index(int, int);
int  high_index(int, int);
void print_force(char *, int);
void print_vel_delta(char *, int);
void print_velocity(char *, int);
void print_tmpA(char *, int);
void shm_reduction(float *, float *);
unsigned align_addr(void *);
int  set_indirection_array(int, int, int);
int  lcm(int, int);	
void print_number_sharers();
void verify(char *, int, float, float);

#ifdef USE_ACK_UPDATES
void mark_nodes_for_update(float *, int);
void send_node_updates(float *, int);
#endif

void bulk_reduction(float *, float *);
void coordinate_bisection(int, double *, double *, double *, char *);

char *global_malloc_sbrk(int, char *);
char *global_malloc(int, char *);
char *local_malloc(int);
char *local_malloc_z(int);

void assign_nodes();
void assign_bnd_nodes();
void assign_edges();
void assign_faces(), assign_bnd_faces();

#ifdef DIRECT_REDUCTION
void init_direct_edge_reduction();
void init_direct_edge_face_reduction();
void direct_reduction(float *, float *, int);
int zero_reduction(float *, int);
#endif

#ifdef FINE_UPDATES
void init_fine_edge_update();
void send_fine_updates(float *, int);
#endif

#ifdef MEMORY_UPDATES

#define TMPA_MODE	0
#define TMPB_MODE	1	/* actually this is not needed */
#define VEL_DELTA_MODE  2
#define VELOCITY_MODE	3

EXTERN int cache_mode[4], home_mode[4], i;

void capture_sharing_list(unsigned, unsigned);
void make_writable_baf_handlers(int);
void set_memory_channels();
void send_memory_updates(unsigned, unsigned);
#endif

#ifdef STACHE_FAULT_INFO
#define MAX_STACHE_FAULT_INFO	2000

EXTERN int stache_fault_info_ctr;

struct stache_fault_info_t
{
    char name[100];
    unsigned startva;
    unsigned endva;
    unsigned num_home_pages;
    int num;
    int num_blks_touched;
};

EXTERN struct stache_fault_info_t stache_fault_info[MAX_STACHE_FAULT_INFO];

void print_stache_fault_info();
void print_debug_stache_fault_info();
void add_stache_fault_info(unsigned va);
void init_stache_fault(char *symbol, unsigned startva, unsigned bytes);
void register_unstructured_page_fault_handler();

#endif STACHE_FAULT_INFO

#ifdef DUMP_MESH_STATS
void dump_mesh_stats();
#endif

#endif _unstructured_h
