/* Copyright (c) 1995 by Mark Hill, James Larus, and David Wood for 
 * the Wisconsin Wind Tunnel Project and Joel Saltz for the High Performance 
 * Software Laboratory, University of Maryland, College Park.
 *
 * ALL RIGHTS RESERVED.
 *
 * This software is furnished under a license and may be used and
 * copied only in accordance with the terms of such license and the
 * inclusion of the above copyright notice.  This software or any other
 * copies thereof or any derivative works may not be provided or
 * otherwise made available to any other persons.  Title to and ownership
 * of the software is retained by Mark Hill, James Larus, Joel Saltz, and
 * David Wood.  Any use of this software must include the above copyright notice.
 *
 * THIS SOFTWARE IS PROVIDED "AS IS".  THE LICENSOR MAKES NO
 * WARRANTIES ABOUT ITS CORRECTNESS OR PERFORMANCE.
 *
 * Shubhendu Mukherjee (Shubu) and Shamik Sharma generated the sequential 
 * version in C. 
 * Shubhendu Mukherjee (Shubu) parallelized unstructured on Tempest. 
 */

/*
 * Joint project between the Wisconsin Wind Tunnel at the University of Wisconsin, 
 * Madison and Joel Saltz's research group at the University of Maryland, College 
 * Park.  
 *
 * Participants: Shubu Mukherjee (*) and Shamik Sharma (!).
 *               Mark Hill (*), Jim Larus(*), Anne Rogers (%), Joel Saltz (!).
 *		 
 *		 (*) University of Wisconsin-Madison
 *		 (!) University of Maryland, College Park
 *		 (%) Princeton University
 */

/* TMPA */
#if (defined(USE_ACK_UPDATES) && !defined(SHM_REDUCTION))
#ifdef RENAME_NODES 
#define send_tmpA_updates() \
          send_node_updates((float *)((unsigned)(TMPA)  \
				      + start_address(myproc)), \
			    high_node - low_node + 1)
#else
#define send_tmpA_updates() \
          send_node_updates((float *)((unsigned)(TMPA) \
				      + start_address(myproc)), \
			    node_list[myproc].num)
#endif RENAME_NODES

#else
#ifdef FINE_UPDATES
#define send_tmpA_updates() \
          send_fine_updates(TMPA, TYPE_EDGE)

#else

#ifdef MEMORY_UPDATES
#ifdef RENAME_NODES 
#define send_tmpA_updates() \
{ \
  if (update_flag) \
    { \
        int first_node = low_node; \
	int tot_nodes  = high_node - low_node + 1; \
	send_memory_updates((unsigned)(&tmpA(first_node,1)), \
			    ((unsigned)(&tmpA(first_node,1)) + \
			     SIZE_OF_NODE * tot_nodes)); \
    } \
}
#else 
#define send_tmpA_updates() \
{ \
  if (update_flag) \
    { \
	int first_node = node_list[myproc].nodes[0]; \
	int tot_nodes  = node_list[myproc].num; \
	send_memory_updates((unsigned)(&tmpA(first_node,1)), \
			    ((unsigned)(&tmpA(first_node,1)) + \
			     SIZE_OF_NODE * tot_nodes)); \
  } \
}
#endif RENAME_NODES

#else
#define send_tmpA_updates()
#endif MEMORY_UPDATES
#endif FINE_UPDATES
#endif USE_ACK_UPDATES


#if (defined(USE_ACK_UPDATES) && !defined(SHM_REDUCTION))
#ifdef RENAME_NODES 
#define mark_tmpA_nodes_for_update() \
         mark_nodes_for_update((float *)((unsigned)(TMPA) \
					 + start_address(myproc)), \
                               (high_node - low_node + 1))
#else 
#define mark_tmpA_nodes_for_update() \
         mark_nodes_for_update((float *)((unsigned)(TMPA) \
					 + start_address(myproc)), \
                               node_list[myproc].num)
#endif RENAME_NODES
#else
#define mark_tmpA_nodes_for_update()
#endif

#ifdef MEMORY_UPDATES
#ifdef RENAME_NODES 
#define capture_tmpA_sharers(update_flag) \
{ \
    if (!(update_flag)) \
      { \
	  int first_node = low_node; \
	  int tot_nodes  = high_node - low_node + 1; \
	  capture_sharing_list((unsigned)(&tmpA(first_node,1)), \
			       (unsigned)(&tmpA(first_node,1)) + \
				SIZE_OF_NODE * tot_nodes); \
	  make_writable_baf_handlers(TMPA_MODE); \
      } \
}
#else 
#define capture_tmpA_sharers(update_flag) \
{ \
    if (!(update_flag)) \
      { \
	  int first_node = node_list[myproc].nodes[0]; \
	  int tot_nodes  = node_list[myproc].num; \
	  capture_sharing_list((unsigned)(&tmpA(first_node,1)), \
			       (unsigned)(&tmpA(first_node,1)) \
			       + SIZE_OF_NODE * tot_nodes); \
	  make_writable_baf_handlers(TMPA_MODE); \
      } \
}
#endif RENAME_NODES
#else
#define capture_tmpA_sharers(update_flag)
#endif


#ifdef MEMORY_UPDATES
#ifdef RENAME_NODES 
#define capture_velocity_sharers(update_flag) \
{ \
    if (!(update_flag)) \
      { \
	  int first_node = low_node; \
	  int tot_nodes  = high_node - low_node + 1; \
	  capture_sharing_list((unsigned)(&velocity(first_node,1)), \
			       ((unsigned)(&velocity(first_node,1)) + \
				SIZE_OF_NODE * tot_nodes)); \
	  make_writable_baf_handlers(VELOCITY_MODE); \
      } \
}
#else
#define capture_velocity_sharers(update_flag) \
{ \
    if (!(update_flag)) \
      { \
	  int first_node = node_list[myproc].nodes[0]; \
	  int tot_nodes  = node_list[myproc].num; \
	  capture_sharing_list((unsigned)(&velocity(first_node,1)), \
			       ((unsigned)(&velocity(first_node,1)) + \
				SIZE_OF_NODE * tot_nodes)); \
	  make_writable_baf_handlers(VELOCITY_MODE); \
      } \
}
#endif RENAME_NODES
#else
#define capture_velocity_sharers(update_flag)	  
#endif

#ifdef FINE_UPDATES
#define send_velocity_updates() send_fine_updates(VELOCITY, TYPE_EDGE_FACE)
#else
#ifdef MEMORY_UPDATES
#ifdef RENAME_NODES 
#define send_velocity_updates() \
{ \
    int first_node = low_node; \
    int tot_nodes  = high_node - low_node + 1; \
    send_memory_updates((unsigned)(&velocity(first_node,1)), \
			((unsigned)(&velocity(first_node,1)) + \
			 SIZE_OF_NODE * tot_nodes)); \
}
#else 
#define send_velocity_sharers() \
{ \
    int first_node = node_list[myproc].nodes[0]; \
    int tot_nodes  = node_list[myproc].num; \
    send_memory_updates((unsigned)(&velocity(first_node,1)), \
			((unsigned)(&velocity(first_node,1)) + \
			 SIZE_OF_NODE * tot_nodes)); \
}
#endif RENAME_NODES
#else
#define send_velocity_updates()
#endif MEMORY_UPDATES
#endif FINE_UPDATES

/* VEL_DELTA */

#ifdef FINE_UPDATES
#define send_vel_delta_updates() \
      send_fine_updates(VEL_DELTA, TYPE_EDGE);
#else
#ifdef MEMORY_UPDATES
#ifdef RENAME_NODES 
#define send_vel_delta_updates() \
{ \
      if (update_flag) \
	{  \
	  int first_node = low_node; \
	  int tot_nodes  = high_node - low_node + 1; \
	  send_memory_updates((unsigned)(&vel_delta(first_node,1)), \
			      (unsigned)(&vel_delta(first_node,1)) + SIZE_OF_NODE * tot_nodes); \
	} \
} 
#else 
#define send_vel_delta_updates() \
{\
      if (update_flag) \
	{\
          int first_node = node_list[myproc].nodes[0]; \
	  int tot_nodes  = node_list[myproc].num; \
	  send_memory_updates((unsigned)(&vel_delta(first_node,1)), \
			      (unsigned)(&vel_delta(first_node,1)) + SIZE_OF_NODE * tot_nodes); \
	} \
} 
#endif RENAME_NODES
#else
#define send_vel_delta_updates()
#endif MEMORY_UPDATES
#endif FINE_UPDATES


#ifdef MEMORY_UPDATES
#ifdef RENAME_NODES 
#define capture_vel_delta_sharers(update_flag, n) \
{ \
      if (!(update_flag) && (n) == 1) \
	{ \
	  int first_node = low_node; \
	  int tot_nodes  = high_node - low_node + 1; \
	  capture_sharing_list((unsigned)(&vel_delta(first_node,1)), \
			      ((unsigned)(&vel_delta(first_node,1)) + \
			       SIZE_OF_NODE * tot_nodes)); \
	  make_writable_baf_handlers(VEL_DELTA_MODE); \
	} \
}
#else 
#define capture_vel_delta_sharers(update_flag, n) \
{ \
      if (!(update_flag) && (n) == 1) \
	{ \
	  int first_node = low_node; \
	  int tot_nodes  = high_node - low_node + 1; \
	  int first_node = node_list[myproc].nodes[0]; \
	  int tot_nodes  = node_list[myproc].num; \
	  capture_sharing_list((unsigned)(&vel_delta(first_node,1)), \
			      ((unsigned)(&vel_delta(first_node,1)) + \
			       SIZE_OF_NODE * tot_nodes)); \
	  make_writable_baf_handlers(VEL_DELTA_MODE); \
	} \
}
#endif RENAME_NODES
#else
#define capture_vel_delta_sharers(update_flag, n)
#endif


